// @(#) ENIB/LI2 (c) - oRis - Version 2.3
//--------------------------------------------------------------------------//
// file        : model.h                                                    //
// description : header file for the ENIB/LI2 class Model                   //
// author(s)   : F. Harrouet                                                //
// copyright   : Laboratoire d'Informatique Industrielle (LI2)              //
//               Ecole Nationale d'Ingenieurs de Brest (ENIB)               //
//               CP 30815 - 29608 Brest Cedex - FRANCE                      //
//               Tel: (33) 02.98.05.66.31 , Fax: (33) 02.98.05.66.29        //
//               e-mail: li2@enib.fr                                        //
//--------------------------------------------------------------------------//
// Copyright (c) 1996-2000                                                  //
// Laboratoire d'Informatique Industrielle (LI2)                            //
// Ecole Nationale d'Ingenieurs de Brest (ENIB)                             //
//                                                                          //
// Permission to use, copy, modify, distribute and sell this software       //
// and its documentation for any purpose is hereby granted without fee,     //
// provided that the above copyright notice appear in all copies and        //
// that both that copyright notice and this permission notice appear        //
// in supporting documentation.                                             //
// ENIB/LI2 makes no representations about the suitability of this          //
// software for any purpose.                                                //
// It is provided "as is" without express or implied warranty.              //
//------------------------------------------------------------------------- //

#ifndef MODEL_H
#define MODEL_H

#include "oRis/Main/agent.h"

#include "Utility/stlList.h"
#include "Utility/stlMap.h"
#include "Utility/stlSet.h"

//---------------------------------------------------------------------------//
// class Model                                                               //
//---------------------------------------------------------------------------//

class Model;

class BroadcastInfo : public BlockAllocatedObject
{
friend class Model;
protected:
  Model * _msgType;
  Agent * _sensitive;
  Word _method;
  BroadcastInfo(Model * msgType,Agent * sensitive,const Word & method)
   : BlockAllocatedObject(),
     _msgType(msgType), _sensitive(sensitive), _method(method) {}
  BroadcastInfo(const BroadcastInfo & i)
   : BlockAllocatedObject(),
     _msgType(i._msgType), _sensitive(i._sensitive), _method(i._method) {}
};

typedef StlList<BroadcastInfo *> BroadcastList;

typedef SimpleVector<Model *> ModelVect;

class Model : public BlockAllocatedObject
{
public:
// Static methods
  inline static Model * getBaseModel(void);
// Allocators
  Model(const Word & name);
  virtual ~Model(void);
// Inspectors
  inline const Word & getName(void) const;
  inline int getRank(void) const;
  inline bool isNative(void) const;
  inline const ModelVect & getParents(void) const;
  bool hasParent(const Model * model) const;
  bool hasParent(const Word & name) const;
  inline const ModelVect & getAncestors(void) const;
  bool hasAncestor(const Model * model) const;
  bool hasAncestor(const Word & name) const;
  inline const ModelVect & getChildren(void) const;
  bool hasChildren(const Model * model) const;
  bool hasChildren(const Word & name) const;
  inline const ModelVect & getDescendants(void) const;
  bool hasDescendant(const Model * model) const;
  bool hasDescendant(const Word & name) const;
  inline bool isA(const Model * model) const;
  bool isA(const Word & name) const;
  Error * tryToSetParent(const Model * parent) const;
  inline long getInstNum(void) const;
  inline int getNbAgents(void) const;
  inline const AgentMap & getAgentMap(void) const;
  inline int getNbAttributes(void) const;
  inline const AttrMap & getAttrMap(void) const;
  const Attribute * getAttribute(const Word & name) const;
  Error * tryToSetAttribute(const Attribute & attr) const;
  inline int getNbModules(void) const;
  inline const ExecModMap & getModuleMap(void) const;
  inline const ExecModule * getMainModule(void) const;
  inline const ExecModule * getNewModule(void) const;
  inline const ExecModule * getDeleteModule(void) const;
  const ExecModule * getModule(const Word & name) const;
  Error * tryToSetModule(const ExecModule * newModule) const;
// Comparisons
  inline bool operator==(const Model & model) const;
  inline bool operator!=(const Model & model) const;
  inline bool operator<(const Model & model) const;
  inline bool operator<=(const Model & model) const;
  inline bool operator>=(const Model & model) const;
  inline bool operator>(const Model & model) const;
// Input/output
  friend ostream& operator<<(ostream& stream, const Model& model);
  Strings printDeclaration(int column=0) const;
  Strings printModules(int column=0) const;
// Modifiers
  inline void setNative(void);
  inline ModelVect & accessParents(void);
  inline ModelVect & accessAncestors(void);
  inline ModelVect & accessChildren(void);
  inline ModelVect & accessDescendants(void);
  void setParent(Model * parent,bool notify);
  inline AgentMap & accessAgentMap(void);
  void setAttribute(const Attribute & attr,bool notify);
  inline AttrMap & accessAttrMap(void);
  Attribute * accessAttribute(const Word & name);
  void setModule(ExecModule * newModule,bool notify);
  inline ExecModMap & accessModuleMap(void);
  inline ExecModule * accessMainModule(void);
  inline ExecModule * accessNewModule(void);
  inline ExecModule * accessDeleteModule(void);
  ExecModule * accessModule(const Word & name);
// Agent Creation/Deletion
  Agent * createAgent(long instNum=0);
  void removeAgent(Agent * agent);
// Broadcast
  void getSensitivity(Agent * sensitive,
                      ModelVect & messages,StringsVect & methods);
  void setSensitivity(Agent * sensitive,const Word & method);
  void cancelSensitivity(Agent * sensitive);
  void removeSensitive(Agent * sensitive);
  void broadcast(Agent * message);
protected:
  Error * _tryToSetAttribute(const Attribute & attr) const;
  Error * _tryToSetModule(const ExecModule * newModule) const;
  bool _setParent(Model * level,Model * parent);
  bool _setAttribute(const Attribute & attr);
  bool _setModule(ExecModule * newModule);
  void _notifyChange(const Word & method,const Word & param);
  BroadcastList::iterator _findBroadcastInfo(Agent * sensitive);
// Attributes
  Word _name;
  int _rank;
  bool _native;
  long _instNum;
  ModelVect _parents;
  ModelVect _ancestors;
  ModelVect _children;
  ModelVect _descendants;
  AgentMap _agentMap;
  AttrMap _attrMap;
  ExecModMap _moduleMap;
  ExecModule * _mainModule;
  ExecModule * _newModule;
  ExecModule * _deleteModule;
  BroadcastList _broadcastList;
  static Model * _baseModel;
  static Word _notifyAncestorWord;
  static Word _notifyAttributeWord;
  static Word _notifyModuleWord;
  static Word _classDescriptionWord;
  static Word _classNameWord;
};

typedef StlMap<Word,Model *> ModelMap;

typedef StlSet<const Model *,StlPtrLess<Model> > ModelSet;

#include "oRis/Main/model.Ci"

#endif // MODEL_H

