// @(#) ENIB/LI2 (c) - oRis - Version 2.3
//--------------------------------------------------------------------------//
// file        : simul.h                                                    //
// description : header file for the ENIB/LI2 class Simulator               //
// author(s)   : F. Harrouet                                                //
// copyright   : Laboratoire d'Informatique Industrielle (LI2)              //
//               Ecole Nationale d'Ingenieurs de Brest (ENIB)               //
//               CP 30815 - 29608 Brest Cedex - FRANCE                      //
//               Tel: (33) 02.98.05.66.31 , Fax: (33) 02.98.05.66.29        //
//               e-mail: li2@enib.fr                                        //
//--------------------------------------------------------------------------//
// Copyright (c) 1996-2000                                                  //
// Laboratoire d'Informatique Industrielle (LI2)                            //
// Ecole Nationale d'Ingenieurs de Brest (ENIB)                             //
//                                                                          //
// Permission to use, copy, modify, distribute and sell this software       //
// and its documentation for any purpose is hereby granted without fee,     //
// provided that the above copyright notice appear in all copies and        //
// that both that copyright notice and this permission notice appear        //
// in supporting documentation.                                             //
// ENIB/LI2 makes no representations about the suitability of this          //
// software for any purpose.                                                //
// It is provided "as is" without express or implied warranty.              //
//--------------------------------------------------------------------------//

#ifndef SIMUL_H
#define SIMUL_H

#include "oRis/Main/model.h"
#include "oRis/Main/parser.h"
#include "oRis/Main/exec.h"
#include "oRis/Main/dLibInfo.h"

#include <signal.h>

//---------------------------------------------------------------------------//
// class Simulator                                                           //
//---------------------------------------------------------------------------//

class SimpleThreadMutex;

class TextOutput;

class Simulator;

typedef Strings (*ControlCB)(Simulator * simul,void * controller,
                             const Strings & command,const Strings & arg);

class Simulator : public BlockAllocatedObject
{
public:
  enum PickingMode {RANDOM, SERIAL};
// Friend functions
  inline friend Simulator * getSimul(void);
// Allocators
  Simulator(const StringsVect & commandLine,
            TextOutput * output,TextOutput * err);
  virtual ~Simulator(void);
// Inspectors
  inline const StringsVect & getPackages(void) const;
  inline bool hasPackage(const Strings & packName) const;
  const SymbolVect * getSymbolVect(const Strings & name) const;
  inline Simulator::PickingMode getPickingMode(void) const;
  inline long getInterval(void) const;
  inline long unsigned int getClock(void) const;
  inline int getMaxDepth(void) const;
  inline const StringsVect & getCommandLine(void) const;
  inline const Strings & getInstallDir(void) const;
  inline const TextOutput * getOutput(void) const;
  inline const TextOutput * getError(void) const;
  inline int getNbErrors(void) const;
  inline bool isBad(void) const;
  inline const StringsVect & getPathVect(void) const;
  Strings locateFile(const Strings & fileName) const;
  inline int getNbAgents(void) const;
  inline const AgentMap & getAgentMap(void) const;
  const Agent * getAgent(const AgentName & name) const;
  inline int getNbModels(void) const;
  inline const ModelMap & getModelMap(void) const;
  const Model * getModel(const Word & name) const;
  inline int getNbModules(void) const;
  inline const ExecModMap & getModuleMap(void) const;
  const ExecModule * getModule(const Word & name) const;
  Error * tryToSetModule(const ExecModule * newModule) const;
  virtual ExternFunction getExternFunction(const Strings & name) const;
  virtual Agent * getNewAgent(Model * model,long instNum) const;
  inline const Execution * getPreviousExecution(const Execution *) const;
// Comparisons
// Input/output
  friend ostream & operator<<(ostream & stream,const Simulator & simul);
  Strings print(void) const;
  bool save(const Strings & fileName) const;
  Strings printStack(void) const;
  virtual void readOrisPath(void);
  void readOrisEdit(void);
  void editFile(const Strings & fileName) const;
  istream * openFile(Strings & fileName) const;
  virtual void showURL(const Strings & url,bool newWindow) const;
// Modifiers
  inline void addPackage(const Strings & packName);
  inline void setControlCB(ControlCB controlCB,void * controller);
  void replaceSymbol(const Strings & name,
                     const SymbolVect * symbolVect=(const SymbolVect *)0);
  inline void setPickingMode(Simulator::PickingMode pickingMode);
  virtual void setInterval(long interval);
  inline void setClock(long unsigned int clock);
  inline void setMaxDepth(int maxDepth);
  inline TextOutput * accessOutput(void);
  inline TextOutput * accessError(void);
  virtual void addToPath(const Strings & path);
  Agent * accessAgent(const AgentName & name);
  void setModel(Model * newModel,bool notify);
  inline ModelMap & accessModelMap(void);
  Model * accessModel(const Word & name);
  void setModule(ExecModule * newModule,bool notify);
  inline ExecModMap & accessModuleMap(void);
  ExecModule * accessModule(const Word & name);
  inline Execution * accessPreviousExecution(const Execution *);
// Simulation Control
  virtual bool openDynLibrary(const Strings & libName);
  int parseFile(Strings & fileName);
  int parseStrings(const Strings & program);
  Value * parseEval(const Strings & program);
  bool parseConstant(Value * result,const Strings & constant);
  void manageAgent(Agent * agent);
  void unManageAgent(Agent * agent);
  void suspendAgent(Agent * agent);
  void resumeAgent(Agent * agent);
  void abortExecution(long execID);
  void suspendExecution(long execID);
  bool resumeExecution(long execID);
  bool isSuspended(long execID);
  long startModule(ExecModule * module,Agent * contextOwner,
                   Execution * initExec=(Execution *)0);
  inline long executeModule(ExecModule * module,Agent * contextOwner,
                            Execution * initExec=(Execution *)0);
  void run(int nbSteps);
  void abortSignal(void);
  void manageError(Error * error);
// Agent Creation/Deletion
  virtual void createAgent(Value * created,bool isTransient,
                           Model * model,long instNum);
  virtual void deleteAgent(Agent * agent);
  bool tryToRemoveExecution(Execution * execution,Agent * agent);
  virtual void removeAgent(Agent * agent);
  virtual void removeDeletedAgent(Agent * agent);
// Utilities
  static void schedHandler(int sigNum);
  inline static unsigned int maxRand(void);
  inline static unsigned int integerRand(unsigned int i);
  inline static double realRand(void);
  inline static double cartToAngle(double x,double y);
  inline static double cartToDist(double x,double y);
  inline static double polToX(double d,double a);
  inline static double polToY(double d,double a);
  inline static void cartToPol(double & d,double & a,double x,double y);
  inline static void polToCart(double & x,double & y,double d,double a);
  inline static double betweenPi(double a);
  static void embedThread(void (* task)(Context *),Context * context);
// Native code

  // native void onExit(string command);
  //
  static void stOnExit(Context * context);

  // native bool loadLibrary(string libName);
  //
  static void stLoadLibrary(Context * context);

  // native int parse(string program);
  //
  static void stParse(Context * context);

  // native int load(string fileName);
  //
  static void stLoad(Context * context);

  // native string[] parseEval(string program);
  //
  static void stParseEval(Context * context);

  // native int eval(string code);
  //
  static void stEval(Context * context);

  // native void save(string fileName);
  //
  static void stSave(Context * context);

  // native int getPickingMode(void);
  //
  static void stGetPickingMode(Context * context);

  // native void setPickingMode(int mode);
  //
  static void stSetPickingMode(Context * context);

  // native int getInterval(void);
  //
  static void stGetInterval(Context * context);

  // native void setInterval(int interval);
  //
  static void stSetInterval(Context * context);

  // native void yield(void);
  //
  static void stYield(Context * context);

  // native void softYield(void);
  //
  static void stSoftYield(Context * context);

  // native void getExecutionStack(Object[] & callers,string[] & modules,
  //                               int execution);
  //
  static void stGetExecutionStack(Context * context);

  // native void printStack(void);
  //
  static void stPrintStack(Context * context);

  // native int currentExecution(void);
  //
  static void stCurrentExecution(Context * context);

  // native void abort(int execution);
  //
  static void stAbort(Context * context);

  // native void suspend(int execution);
  //
  static void stSuspend(Context * context);

  // native void resume(int execution);
  //
  static void stResume(Context * context);

  // native bool isSuspended(int execution);
  //
  static void stIsSuspended(Context * context);

  // native int getMaxDepth(void);
  //
  static void stGetMaxDepth(Context * context);

  // native void setMaxDepth(int maxDepth);
  //
  static void stSetMaxDepth(Context * context);

  // native int getClock(void);
  //
  static void stGetClock(Context * context);

  // native void setClock(int clk); // not documented
  //
  static void stSetClock(Context * context);

  // native Object create(string objName); // not documented
  //
  static void stCreate(Context * context);

  // native Object getObject(string objName);
  //
  static void stGetObject(Context * context);

  // native bool exist(Object obj);
  //
  static void stExist(Context * context);

  // native int numberOf(string className);
  //
  static void stNumberOf(Context * context);

  // native Object[] getAll(string className);
  //
  static void stGetAll(Context * context);

  // native string[] getClasses(void);
  //
  static void stGetClasses(Context * context);

  // native int getTime(void);
  //
  static void stGetTime(Context * context);

  // native string getenv(string varName);
  //
  static void stGetenv(Context * context);

  // native void putenv(string varName,string value);
  //
  static void stPutenv(Context * context);

  // native string systemName(bool detailed);
  //
  static void stSystemName(Context * context);

  // native void systemYield(void);
  //
  static void stSystemYield(Context * context);

  // native int system(string command);
  //
  static void stSystem(Context * context);

  // native void systemI(string command,string input);
  //
  static void stSystemI(Context * context);

  // native string systemO(string command);
  //
  static void stSystemO(Context * context);

  // native string systemE(string command);
  //
  static void stSystemE(Context * context);

  // native string systemIO(string command,string input);
  //
  static void stSystemIO(Context * context);

  // native string systemIE(string command,string input);
  //
  static void stSystemIE(Context * context);

  // native string systemOE(string command,string & err);
  //
  static void stSystemOE(Context * context);

  // native string systemIOE(string command,string input,string & err);
  //
  static void stSystemIOE(Context * context);

  // native string control(string command,string argument);
  //
  static void stControl(Context * context);

  // native string[] commandLine(bool full);
  //
  static void stCommandLine(Context * context);

  // native int asciiCode(string char);
  //
  static void stAsciiCode(Context * context);

  // native string asciiChar(int code);
  //
  static void stAsciiChar(Context * context);

  // native string hexaString(int value);
  //
  static void stHexaString(Context * context);

  // native string octalString(int value);
  //
  static void stOctalString(Context * context);

  // native string binaryString(int value);
  //
  static void stBinaryString(Context * context);

  // native string locateFile(string fileName);
  //
  static void stLocateFile(Context * context);

  // native void showURL(string url,bool newWindow);
  //
  static void stShowURL(Context * context);

  // native string readln(void);
  //
  static void stReadln(Context * context);

protected:
  static void _systemTask(Context * context);
  static void _systemITask(Context * context);
  static void _systemOTask(Context * context);
  static void _systemETask(Context * context);
  static void _systemIOTask(Context * context);
  static void _systemIETask(Context * context);
  static void _systemOETask(Context * context);
  static void _systemIOETask(Context * context);
  static void _readlnTask(Context * context);
  void _notifyChange(const Word & method,const Word & param);
// Attributes
  DynLibInfoVect _dynLibs;
  StringsVect _packages;
  SymbolMap _replaceMap;
  Simulator::PickingMode _pickingMode;
  long _interval;
  unsigned long _clock;
  SimpleThreadMutex * _readlnMutex;
  int _maxDepth;
  StringsVect _commandLine;
  Strings _installDir;
  TextOutput * _output;
  TextOutput * _err;
  int _nbErrors;
  bool _isBad;
  StringsVect _pathVect;
  AgentMap _agentMap;
  AgentMap _deletedMap;
  ExecutionVect _execVectFrom;
  ExecutionVect _execVectTo;
  ExecutionList _execListSuspend;
  ExecutionVect _curExecutions;
  ModelMap _modelMap;
  ExecModMap _moduleMap;
  ValueVect _parseResults;
  Strings _editCommand;
  SimpleVector<Strings *> _onExit;
  ControlCB _controlCB;
  void * _controller;
  static Simulator * _simul;
  static struct sigaction _newTimerAction;
  static struct sigaction _oldTimerAction;
  static Word _notifyClassWord;
  static Word _notifyModuleWord;
  static Word _globalDescriptionWord;
  static const char * _launchCode1; // one string is too long under MSDOS !!!
  static const char * _launchCode2;
  static const char * _launchCode3;
};

#include "oRis/Main/simul.Ci"

#endif // SIMUL_H

