// @(#) ENIB/LI2 (c) - oRis - Version 2.3
//--------------------------------------------------------------------------//
// file        : value.h                                                    //
// description : header file for the ENIB/LI2 class Value                   //
// author(s)   : F. Harrouet & J. Tisseau                                   //
// copyright   : Laboratoire d'Informatique Industrielle (LI2)              //
//               Ecole Nationale d'Ingenieurs de Brest (ENIB)               //
//               CP 30815 - 29608 Brest Cedex - FRANCE                      //
//               Tel: (33) 02.98.05.66.31 , Fax: (33) 02.98.05.66.29        //
//               e-mail: li2@enib.fr                                        //
//--------------------------------------------------------------------------//
// Copyright (c) 1996-2000                                                  //
// Laboratoire d'Informatique Industrielle (LI2)                            //
// Ecole Nationale d'Ingenieurs de Brest (ENIB)                             //
//                                                                          //
// Permission to use, copy, modify, distribute and sell this software       //
// and its documentation for any purpose is hereby granted without fee,     //
// provided that the above copyright notice appear in all copies and        //
// that both that copyright notice and this permission notice appear        //
// in supporting documentation.                                             //
// ENIB/LI2 makes no representations about the suitability of this          //
// software for any purpose.                                                //
// It is provided "as is" without express or implied warranty.              //
//--------------------------------------------------------------------------//

#ifndef VALUE_H
#define VALUE_H

#include "oRis/Main/constant.h"
#include "Utility/word.h"

//--------------------------------------------------------------------------//
// class ValueType                                                          //
//--------------------------------------------------------------------------//

class Model;
class Agent;
class Context;

class ValueType : public BlockAllocatedObject
{
public:
  enum SimpleType {UNKNOWN, INTEGER, REAL, STRING, AGENT};
// Allocators
  inline ValueType(SimpleType type=ValueType::UNKNOWN,int depth=0);
  inline ValueType(Model * model,int depth=0);
  inline ValueType(const ValueType & type);
  inline virtual ~ValueType(void);
// Inspectors
  inline ValueType::SimpleType getSimpleType(void) const;
  inline Model * getModel(void) const;
  inline int getDepth(void) const;
  bool canCastFrom(const ValueType & from) const;
  inline bool canCastTo(const ValueType & to) const;
// Comparisons
  inline bool operator==(const ValueType & type) const;
  inline bool operator!=(const ValueType & type) const;
// Input/output
  friend ostream & operator<<(ostream & stream,const ValueType & type);
  Strings print(void) const;
// Modifiers
  inline void setSimpleType(ValueType::SimpleType type);
  inline void setModel(Model * model);
  inline Model * accessModel(void);
  inline void setDepth(int depth);
  inline ValueType & operator=(const ValueType & type);
private:
// Attributes
  ValueType::SimpleType _type;
  Model * _model;
  int _depth;
};

//--------------------------------------------------------------------------//
// class Value                                                              //
//--------------------------------------------------------------------------//

class Value;
typedef SimpleVector<Value *> ValueVect;

class SubValue;
class SubValueVect : public SimpleVector<SubValue *>
{
public:
  friend class SubValue;
protected:                                  // use static methods of SubValue
  inline SubValueVect(unsigned int size=0)
          : SimpleVector<SubValue *>(size) {}
  inline SubValueVect(const SubValueVect & vect)
          : SimpleVector<SubValue *>(vect) {}
  inline virtual ~SubValueVect(void) {}
};

class Value : public BlockAllocatedObject
{
public:
// Allocators
  inline Value(void);
  inline Value(const ValueType & type);
  Value(const Value & value);
  virtual ~Value(void);
// Inspectors
  inline virtual bool isLeftValue(void) const;
  inline long getInteger(void) const;
  inline double getReal(void) const;
  inline const Strings & getStrings(void) const;
  inline Agent * getAgent(void) const;
  inline const SubValueVect * getValueVector(void) const;
  inline int getLen(void) const;
  inline const Value & getElem(int pos) const;
  inline const Value & getFront(void) const;
  inline const Value & getBack(void) const;
  int findElem(const Value & value,bool deep) const;
  int minPosition(void) const;
  int maxPosition(void) const;
  bool isNull(void) const;
  inline const ValueType & getType(void) const;
// Comparisons
  bool operator==(const Value & value) const;
  bool deepEquals(const Value & value) const;
  inline bool operator!=(const Value & value) const;
  bool operator<(const Value & value) const;
  inline bool operator>(const Value & value) const;
  inline bool operator>=(const Value & value) const;
  inline bool operator<=(const Value & value) const;
// Input/output
  friend ostream & operator<<(ostream & stream,const Value & value);
  Strings print(void) const;
  Strings dump(void) const;
// Modifiers
  inline virtual void reference(void);
  inline virtual void relax(void);
  virtual void clearReferences(bool strongClear=false);
  inline void setType(const ValueType & type);
  inline ValueType & accessType(void);
  inline virtual bool preset(void);
  inline virtual void postset(void);
  void razVector(void);
  void copyVector(const SubValueVect & valueVector);
  inline virtual void setInteger(long integer);
  inline virtual void setReal(double real);
  inline virtual void setStrings(const Strings & string);
  virtual void setAgent(Agent * agent);
  inline virtual void setValueVectorCopy(const Value & valueVector);
  virtual void setValueVectorCopyI2R(const SubValueVect & valueVector);
  virtual void setValueVectorCopyR2I(const SubValueVect & valueVector);
  inline virtual void setValueVector(SubValueVect * valueVector);
  inline virtual Value * accessElem(int pos);
  inline virtual Value * accessFront(void);
  inline virtual Value * accessBack(void);
  inline virtual void insert(SubValue * value,int pos);
  inline virtual void insert(const Value & value,int pos);
  inline virtual void pushFront(SubValue * value);
  inline virtual void pushFront(const Value & value);
  inline virtual void pushBack(SubValue * value);
  inline virtual void pushBack(const Value & value);
  inline virtual void erase(int pos);
  inline virtual void readErase(Value & result,int pos);
  inline virtual void popFront(void);
  inline virtual void readPopFront(Value & result);
  inline virtual void popBack(void);
  inline virtual void readPopBack(Value & result);
  virtual void appendVector(const SubValueVect & valueVector);
  bool evalVectType(void);
  virtual void sortVector(void);
  virtual Value & operator=(const Value & value);
  Value & cast(const ValueType & type);
  bool deepCast(const ValueType & type);
  virtual Value & affectCast(const Value & value);
protected:
// Attributes
  ValueType _type;
  long _integer;
  double _real;
  Strings _string;
  Agent * _agent;
  SubValueVect * _valueVector;
  static Word _equalsToWord;
  static Word _lessThanWord;
};

//---------------------------------------------------------------------------//
// class SubValue                                                            //
//---------------------------------------------------------------------------//

class SubValue : public Value
{
public:
// Static methods
  inline static SubValue * getSubValue(const ValueType & type,Value * upper);
  inline static SubValue * getSubValue(const Value & value,Value * upper);
  inline static void relaxSubValue(SubValue * value);
  static void clearSubValueStock(void);
  inline static SubValueVect * getSubValueVect(void);
  inline static void relaxSubValueVect(SubValueVect * vect);
  static void clearSubValueVectStock(void);
// Allocators
protected:                                               // use static methods
  inline SubValue(const ValueType & type,Value * upper);
  inline SubValue(const Value & value,Value * upper);
  inline virtual ~SubValue(void);
public:
// Inspectors
  inline virtual bool isLeftValue(void) const;
  inline const Value * getUpper(void) const;
// Comparisons
// Input/output
// Modifiers
  inline void setUpper(Value * upper);
  inline Value * accessUpper(void);
  inline virtual void reference(void);
  inline virtual void relax(void);
  inline virtual bool preset(void);
  inline virtual void postset(void);
  inline virtual void setInteger(long integer);
  inline virtual void setReal(double real);
  inline virtual void setStrings(const Strings & string);
  inline virtual void setAgent(Agent * agent);
  inline virtual void setValueVectorCopy(const Value & valueVector);
  inline virtual void setValueVectorCopyI2R(const SubValueVect & valueVector);
  inline virtual void setValueVectorCopyR2I(const SubValueVect & valueVector);
  inline virtual void setValueVector(SubValueVect * valueVector);
  inline virtual void insert(SubValue * value,int pos);
  inline virtual void insert(const Value & value,int pos);
  inline virtual void pushFront(SubValue * value);
  inline virtual void pushFront(const Value & value);
  inline virtual void pushBack(SubValue * value);
  inline virtual void pushBack(const Value & value);
  inline virtual void erase(int pos);
  inline virtual void readErase(Value & result,int pos);
  inline virtual void popFront(void);
  inline virtual void readPopFront(Value & result);
  inline virtual void popBack(void);
  inline virtual void readPopBack(Value & result);
  inline virtual void appendVector(const SubValueVect & valueVector);
  inline virtual void sortVector(void);
  inline virtual Value & operator=(const Value & value);
  inline virtual Value & affectCast(const Value & value);
protected:
// Attributes
  Value * _upper;
  static SimpleVector<SubValue *> _subValueStock;
  static SimpleVector<SubValueVect *> _subValueVectStock;
};

#include "oRis/Main/value.Ci"

#endif // VALUE_H

