// @(#) ENIB/LI2 (c) - oRis - Version 2.3
//--------------------------------------------------------------------------//
// file        : utilCd.cpp                                                 //
// description : source file for the ENIB/LI2 Code utilities                //
// author(s)   : F. Harrouet                                                //
// copyright   : Laboratoire d'Informatique Industrielle (LI2)              //
//               Ecole Nationale d'Ingenieurs de Brest (ENIB)               //
//               CP 30815 - 29608 Brest Cedex - FRANCE                      //
//               Tel: (33) 02.98.05.66.31 , Fax: (33) 02.98.05.66.29        //
//               e-mail: li2@enib.fr                                        //
//--------------------------------------------------------------------------//
// Copyright (c) 1996-2000                                                  //
// Laboratoire d'Informatique Industrielle (LI2)                            //
// Ecole Nationale d'Ingenieurs de Brest (ENIB)                             //
//                                                                          //
// Permission to use, copy, modify, distribute and sell this software       //
// and its documentation for any purpose is hereby granted without fee,     //
// provided that the above copyright notice appear in all copies and        //
// that both that copyright notice and this permission notice appear        //
// in supporting documentation.                                             //
// ENIB/LI2 makes no representations about the suitability of this          //
// software for any purpose.                                                //
// It is provided "as is" without express or implied warranty.              //
//--------------------------------------------------------------------------//

#include "oRis/Codes/utilCd.h"
#include "oRis/Main/context.h"
#include "oRis/Main/simul.h"
#include "oRis/Main/textOut.h"
#include "oRis/Main/errors.h"
#include "oRis/Nodes/smaNd.h"

//--------------------------------------------------------------------------//
// class PopCode                                                            //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
PopCode::PopCode(const SmaNode * node)
 : SmaCode(node)
{}

PopCode::~PopCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void PopCode::eval(Context * context) const // virtual
{
context->accessExecution()->popValue()->relax();
}


//--------------------------------------------------------------------------//
// class PushCode                                                           //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
PushCode::PushCode(const SmaNode * node,const ValueType & valueType)
 : SmaCode(node), _valueType(valueType)
{}

PushCode::~PushCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void PushCode::eval(Context * context) const // virtual
{
context->accessExecution()->pushValue(TmpValue::getTmpValue(_valueType));
}


//--------------------------------------------------------------------------//
// class JumpCode                                                           //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
JumpCode::JumpCode(const SmaNode * node,int offset)
 : SmaCode(node), _offset(offset)
{}

JumpCode::~JumpCode(void)
{}

//-------- Inspectors ------------------------------------------------------//
int JumpCode::jumpsTo(void) const // virtual
{
return(_offset);
}

//-------- Evaluation ------------------------------------------------------//
void JumpCode::eval(Context * context) const // virtual
{
context->moveCounter(_offset);
}


//--------------------------------------------------------------------------//
// class JumpIfTrueCode                                                     //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
JumpIfTrueCode::JumpIfTrueCode(const SmaNode * node,int offset)
 : SmaCode(node), _offset(offset)
{}

JumpIfTrueCode::~JumpIfTrueCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void JumpIfTrueCode::eval(Context * context) const // virtual
{
Value * value=context->accessExecution()->popValue();
if(!value->isNull()) context->moveCounter(_offset);
value->relax();
}


//--------------------------------------------------------------------------//
// class JumpIfTrueNoPopCode                                                //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
JumpIfTrueNoPopCode::JumpIfTrueNoPopCode(const SmaNode * node,int offset)
 : SmaCode(node), _offset(offset)
{}

JumpIfTrueNoPopCode::~JumpIfTrueNoPopCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void JumpIfTrueNoPopCode::eval(Context * context) const // virtual
{
if(!context->accessExecution()->lastValue()->isNull())
  context->moveCounter(_offset);
}


//--------------------------------------------------------------------------//
// class JumpIfFalseCode                                                    //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
JumpIfFalseCode::JumpIfFalseCode(const SmaNode * node,int offset)
 : SmaCode(node), _offset(offset)
{}

JumpIfFalseCode::~JumpIfFalseCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void JumpIfFalseCode::eval(Context * context) const // virtual
{
Value * value=context->accessExecution()->popValue();
if(value->isNull()) context->moveCounter(_offset);
value->relax();
}


//--------------------------------------------------------------------------//
// class JumpIfFalseNoPopCode                                               //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
JumpIfFalseNoPopCode::JumpIfFalseNoPopCode(const SmaNode * node,int offset)
 : SmaCode(node), _offset(offset)
{}

JumpIfFalseNoPopCode::~JumpIfFalseNoPopCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void JumpIfFalseNoPopCode::eval(Context * context) const // virtual
{
if(context->accessExecution()->lastValue()->isNull())
  context->moveCounter(_offset);
}


//--------------------------------------------------------------------------//
// class NotCode                                                            //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
NotCode::NotCode(const SmaNode * node)
 : SmaCode(node)
{}

NotCode::~NotCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void NotCode::eval(Context * context) const // virtual
{
Value * value=context->accessExecution()->popValue();
if(value->isNull())
  {
  value->relax();
  value=TmpValue::getTmpValue(_node->getType());
  value->setInteger(1);
  }
else
  {
  value->relax();
  value=TmpValue::getTmpValue(_node->getType());
  value->setInteger(0);
  }
context->accessExecution()->pushValue(value);
}


//--------------------------------------------------------------------------//
// class ReturnCode                                                         //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
ReturnCode::ReturnCode(const SmaNode * node,int noCBLevel,int criticalLevel)
 : SmaCode(node), _noCBLevel(noCBLevel), _criticalLevel(criticalLevel)
{}

ReturnCode::~ReturnCode(void)
{}

//-------- Inspectors ------------------------------------------------------//
bool ReturnCode::isReturn(void) const // virtual
{
return(true);
}

//-------- Evaluation ------------------------------------------------------//
void ReturnCode::eval(Context * context) const // virtual
{
for(int i=_noCBLevel;i--;) ValueCallback::unmask();
if(_criticalLevel) context->accessExecution()->endCritical(_criticalLevel);
context->accessExecution()->leaveContext();
}


//--------------------------------------------------------------------------//
// class BeginCriticalCode                                                  //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
BeginCriticalCode::BeginCriticalCode(const SmaNode * node)
 : SmaCode(node)
{}

BeginCriticalCode::~BeginCriticalCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void BeginCriticalCode::eval(Context * context) const // virtual
{
context->accessExecution()->beginCritical();
}


//--------------------------------------------------------------------------//
// class EndCriticalCode                                                    //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
EndCriticalCode::EndCriticalCode(const SmaNode * node)
 : SmaCode(node)
{}

EndCriticalCode::~EndCriticalCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void EndCriticalCode::eval(Context * context) const // virtual
{
context->accessExecution()->endCritical();
}

//--------------------------------------------------------------------------//
// class BeginNoCBCode                                                      //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
BeginNoCBCode::BeginNoCBCode(const SmaNode * node)
 : SmaCode(node)
{}

BeginNoCBCode::~BeginNoCBCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void BeginNoCBCode::eval(Context *) const // virtual
{
ValueCallback::mask();
}


//--------------------------------------------------------------------------//
// class EndNoCBCode                                                        //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
EndNoCBCode::EndNoCBCode(const SmaNode * node)
 : SmaCode(node)
{}

EndNoCBCode::~EndNoCBCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void EndNoCBCode::eval(Context *) const // virtual
{
ValueCallback::unmask();
}


//--------------------------------------------------------------------------//
// class CopyCode                                                           //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
CopyCode::CopyCode(const SmaNode * node)
 : SmaCode(node)
{}

CopyCode::~CopyCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void CopyCode::eval(Context * context) const // virtual
{
Value * value=context->accessExecution()->popValue();
Value * result=TmpValue::getTmpValue(value->getType());
*result=*value;
value->relax();
context->accessExecution()->pushValue(result);
}

//--------------------------------------------------------------------------//
// class CastCode                                                           //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
CastCode::CastCode(const SmaNode * node,const ValueType & resultType)
 : SmaCode(node), _baseType(resultType), _resultType(resultType)
{
_baseType.setDepth(0);
}

CastCode::~CastCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void CastCode::eval(Context * context) const // virtual
{
Value * value=context->accessExecution()->popValue();
Value * result=TmpValue::getTmpValue(value->getType());
*result=*value;
value->relax();
if(result->deepCast(_resultType))
  {
  context->accessExecution()->pushValue(result);
  return;
  }
Strings typeName=result->getType().print();
result->relax();
context->accessExecution()
       ->pushValue(TmpValue::getTmpValue(_resultType)); // not used
Strings message="Can't cast '"+
                typeName+
                "' to '"+
                _resultType.print()+
                "'";
manageError(new Error(Error::RUNTIME,message));
}


//--------------------------------------------------------------------------//
// class FormatCode                                                         //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
FormatCode::FormatCode(const SmaNode * node,int nbArgs)
 : SmaCode(node), _nbArgs(nbArgs)
{}

FormatCode::~FormatCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void FormatCode::eval(Context * context) const // virtual
{
Value * value;
Strings str;
for(int i=_nbArgs;i--;)
  {
  value=context->accessExecution()->popValue();
  str+=value->print();
  value->relax();
  }
Value * result=TmpValue::getTmpValue(_node->getType());
result->setStrings(str);
context->accessExecution()->pushValue(result);
}


//--------------------------------------------------------------------------//
// class PrintCode                                                          //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
PrintCode::PrintCode(const SmaNode * node,int nbArgs,bool newLine)
 : SmaCode(node), _nbArgs(nbArgs), _newLine(newLine)
{}

PrintCode::~PrintCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void PrintCode::eval(Context * context) const // virtual
{
Value * value;
for(int i=_nbArgs;i--;)
  {
  value=context->accessExecution()->popValue();
  *getSimul()->accessOutput() << value->print();
  value->relax();
  }
if(_newLine) *getSimul()->accessOutput() << endl;
else *getSimul()->accessOutput() << flush;
}


//--------------------------------------------------------------------------//
// class ErrorCode                                                          //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
ErrorCode::ErrorCode(const SmaNode * node,int nbArgs,bool warning)
 : SmaCode(node), _nbArgs(nbArgs), _warning(warning)
{}

ErrorCode::~ErrorCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void ErrorCode::eval(Context * context) const // virtual
{
Strings str;
Value * value;
for(int i=_nbArgs;i--;)
  {
  value=context->accessExecution()->popValue();
  str+=value->print();
  value->relax();
  }
if(_warning) manageError(new Error(Error::WARNING,str));
else
  {
  if(context->getModule()->getResultType().getSimpleType()!=ValueType::UNKNOWN)
    {
    context->returnValue(TmpValue::getTmpValue(context->getModule()
                                                      ->getResultType()));
    }
  manageError(new Error(Error::RUNTIME,str));
  }
}


