// @(#) ENIB/LI2 (c) - oRis - Version 2.3
//--------------------------------------------------------------------------//
// file        : valCd.cpp                                                  //
// description : source file for the ENIB/LI2 Values access codes           //
// author(s)   : F. Harrouet                                                //
// copyright   : Laboratoire d'Informatique Industrielle (LI2)              //
//               Ecole Nationale d'Ingenieurs de Brest (ENIB)               //
//               CP 30815 - 29608 Brest Cedex - FRANCE                      //
//               Tel: (33) 02.98.05.66.31 , Fax: (33) 02.98.05.66.29        //
//               e-mail: li2@enib.fr                                        //
//--------------------------------------------------------------------------//
// Copyright (c) 1996-2000                                                  //
// Laboratoire d'Informatique Industrielle (LI2)                            //
// Ecole Nationale d'Ingenieurs de Brest (ENIB)                             //
//                                                                          //
// Permission to use, copy, modify, distribute and sell this software       //
// and its documentation for any purpose is hereby granted without fee,     //
// provided that the above copyright notice appear in all copies and        //
// that both that copyright notice and this permission notice appear        //
// in supporting documentation.                                             //
// ENIB/LI2 makes no representations about the suitability of this          //
// software for any purpose.                                                //
// It is provided "as is" without express or implied warranty.              //
//--------------------------------------------------------------------------//

#include "oRis/Codes/valCd.h"
#include "oRis/Main/context.h"
#include "oRis/Main/simul.h"
#include "oRis/Main/tmpVal.h"
#include "oRis/Main/errors.h"
#include "oRis/Nodes/smaNd.h"

//--------------------------------------------------------------------------//
// class ConstCode                                                          //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
ConstCode::ConstCode(const SmaNode * node,const Value & value)
 : SmaCode(node), _value(new ConstValue(value.getType()))
{
_value->reference();
_value->affectCast(value);
}

ConstCode::~ConstCode(void)
{
_value->relax();
}

//-------- Evaluation ------------------------------------------------------//
void ConstCode::eval(Context * context) const // virtual
{
_value->reference();
context->accessExecution()->pushValue((Value *)_value);
}


//--------------------------------------------------------------------------//
// class RealRndCode                                                        //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
RealRndCode::RealRndCode(const SmaNode * node)
 : SmaCode(node)
{}

RealRndCode::~RealRndCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void RealRndCode::eval(Context * context) const // virtual
{
Value * value=TmpValue::getTmpValue(_node->getType());
value->setReal(Simulator::realRand());
context->accessExecution()->pushValue(value);
}


//--------------------------------------------------------------------------//
// class IntRnd_I_Code                                                      //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
IntRnd_I_Code::IntRnd_I_Code(const SmaNode * node)
 : SmaCode(node)
{}

IntRnd_I_Code::~IntRnd_I_Code(void)
{}

//-------- Evaluation ------------------------------------------------------//
void IntRnd_I_Code::eval(Context * context) const // virtual
{
Value * value=context->accessExecution()->popValue();
unsigned long max=value->getInteger();
value->relax();
ValueType resultType(_node->getType());
value=TmpValue::getTmpValue(resultType);
if((max>=1)&&(max<getSimul()->maxRand()))
  {
  value->setInteger(Simulator::integerRand(max));
  context->accessExecution()->pushValue(value);
  }
else
  {
  context->accessExecution()->pushValue(value); // not used
  Strings message="Can use '"+
                  _node->getName()+
                  "' only from "+Strings(1)+
                  " to "+Strings(getSimul()->maxRand()-1)+
                  " (called with "+Strings(max)+")";
  manageError(new Error(Error::RUNTIME,message));
  }
}


//--------------------------------------------------------------------------//
// class IntRnd_R_Code                                                      //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
IntRnd_R_Code::IntRnd_R_Code(const SmaNode * node)
 : SmaCode(node)
{}

IntRnd_R_Code::~IntRnd_R_Code(void)
{}

//-------- Evaluation ------------------------------------------------------//
void IntRnd_R_Code::eval(Context * context) const // virtual
{
Value * value=context->accessExecution()->popValue();
unsigned long max=(long)value->getReal();
value->relax();
ValueType resultType(_node->getType());
value=TmpValue::getTmpValue(resultType);
if((max>=1)&&(max<getSimul()->maxRand()))
  {
  value->setInteger(Simulator::integerRand(max));
  context->accessExecution()->pushValue(value);
  }
else
  {
  context->accessExecution()->pushValue(value); // not used
  Strings message="Can use '"+
                  _node->getName()+
                  "' only from "+Strings(1)+
                  " to "+Strings(getSimul()->maxRand()-1)+
                  " (called with "+Strings(max)+")";
  manageError(new Error(Error::RUNTIME,message));
  }
}


//--------------------------------------------------------------------------//
// class SelfCode                                                           //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
SelfCode::SelfCode(const SmaNode * node)
 : SmaCode(node)
{}

SelfCode::~SelfCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void SelfCode::eval(Context * context) const // virtual
{
Value * result=TmpValue::getTmpValue(context->accessOwner()->accessModel());
result->setAgent(context->accessOwner());
context->accessExecution()->pushValue(result);
}


//--------------------------------------------------------------------------//
// class CallerCode                                                         //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
CallerCode::CallerCode(const SmaNode * node)
 : SmaCode(node)
{}

CallerCode::~CallerCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void CallerCode::eval(Context * context) const // virtual
{
Value * result;
Agent * caller=(Agent *)0;
if(context->getModule()->getSpecialType()==ExecModule::CONSTRUCTOR)
  {
  Execution * execution=context->accessExecution();
  execution=getSimul()->accessPreviousExecution(execution);
  if(execution)
    caller=execution->accessContext()->accessOwner();
  }
else
  {
  if(context->getPrevious())
    caller=context->accessPrevious()->accessOwner();
  }
if(caller)
  result=TmpValue::getTmpValue(caller->accessModel());
else
  result=TmpValue::getTmpValue(Model::getBaseModel());
result->setAgent(caller);
context->accessExecution()->pushValue(result);
}


//--------------------------------------------------------------------------//
// class DumpCode                                                           //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
DumpCode::DumpCode(const SmaNode * node)
 : SmaCode(node)
{}

DumpCode::~DumpCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void DumpCode::eval(Context * context) const // virtual
{
Value * result=TmpValue::getTmpValue(_node->getType());
Value * value=context->accessExecution()->popValue();
result->setStrings(value->dump());
value->relax();
context->accessExecution()->pushValue(result);
}


//--------------------------------------------------------------------------//
// class ValueOfCode                                                        //
//--------------------------------------------------------------------------//

//-------- Allocators ------------------------------------------------------//
ValueOfCode::ValueOfCode(const SmaNode * node)
 : SmaCode(node)
{}

ValueOfCode::~ValueOfCode(void)
{}

//-------- Evaluation ------------------------------------------------------//
void ValueOfCode::eval(Context * context) const // virtual
{
Value * value=context->accessExecution()->popValue();
Value * result=context->accessExecution()->popValue();
Value * succes=TmpValue::getTmpValue(ValueType(ValueType::INTEGER));
if(getSimul()->parseConstant(result,value->getStrings()))
  succes->setInteger(1);
else
  succes->setInteger(0);
value->relax();
result->relax();
context->accessExecution()->pushValue(succes);
}

