// @(#) ENIB/LI2 (c) - oRis - Version 2.3
//--------------------------------------------------------------------------//
// file        : context.cpp                                                //
// description : source file for the ENIB/LI2 class Context                 //
// author(s)   : F. Harrouet                                                //
// copyright   : Laboratoire d'Informatique Industrielle (LI2)              //
//               Ecole Nationale d'Ingenieurs de Brest (ENIB)               //
//               CP 30815 - 29608 Brest Cedex - FRANCE                      //
//               Tel: (33) 02.98.05.66.31 , Fax: (33) 02.98.05.66.29        //
//               e-mail: li2@enib.fr                                        //
//--------------------------------------------------------------------------//
// Copyright (c) 1996-2000                                                  //
// Laboratoire d'Informatique Industrielle (LI2)                            //
// Ecole Nationale d'Ingenieurs de Brest (ENIB)                             //
//                                                                          //
// Permission to use, copy, modify, distribute and sell this software       //
// and its documentation for any purpose is hereby granted without fee,     //
// provided that the above copyright notice appear in all copies and        //
// that both that copyright notice and this permission notice appear        //
// in supporting documentation.                                             //
// ENIB/LI2 makes no representations about the suitability of this          //
// software for any purpose.                                                //
// It is provided "as is" without express or implied warranty.              //
//--------------------------------------------------------------------------//

#include "oRis/Main/context.h"
#include "oRis/Main/tmpVal.h"
#include "oRis/Main/agent.h"
#include "oRis/Main/exec.h"
#include "oRis/Main/errors.h"
#include "oRis/Nodes/smaNd.h"

//--------------------------------------------------------------------------//
// class Context                                                            //
//--------------------------------------------------------------------------//

SimpleVector<Context *> Context::_contextStock;
SimpleVector<ValueVect *> Context::_valueVectStock;

//-------- Static methods --------------------------------------------------//
Context * Context::getContext(ExecModule * module,Agent * owner,
                              Context * previous) // static
{
Context * context;
if(Context::_contextStock.getSize())
  {
  context=Context::_contextStock.readPopBack();
  }
else context=new Context;
context->_module=module;
context->_code=module->getCodeVect();
context->_owner=owner;
context->_previous=previous;
return(context);
}

// inline void Context::relaxContext(Context * context) // static

void Context::clearContextStock(void) // static
{
while(Context::_contextStock.getSize())
  {
  delete Context::_contextStock.readPopBack();
  }
}

// inline ValueVect * Context::getValueVect(void) // static

// inline void Context::relaxValueVect(ValueVect * vect) // static

void Context::clearValueVectStock(void) // static
{
while(Context::_valueVectStock.getSize())
  {
  delete Context::_valueVectStock.readPopBack();
  }
}

//-------- Allocators -------------------------------------------------------//
Context::Context(void)
 : BlockAllocatedObject(),
   _execution((Execution *)0), _counter(0),
   _module((ExecModule *)0), _locals((ValueVect *)0),
   _owner((Agent *)0), _code((const CodeVect *)0),
   _previous((Context *)0)
{}

Context::~Context(void)
{}

//-------- Inspectors ------------------------------------------------------//
// inline const Execution * Context::getExecution(void) const

// inline const Context * Context::getPrevious(void) const

// inline const Agent * Context::getOwner(void) const

// inline const ExecModule * Context::getModule(void) const

// inline const ValueVect * Context::getLocals(void) const

// inline const Value * Context::getLocal(int rank) const

// inline long Context::getInteger(int rank) const

// inline double Context::getReal(int rank) const

// inline const Strings & Context::getStrings(int rank) const

// inline Agent * Context::getAgent(int rank) const

// inline const SubValueVect & Context::getValueVector(int rank) const

void Context::getStack(AgentVect & callers,StringsVect & modules,
                       int depth) const
{
if(depth)
  {
  if(_previous) _previous->getStack(callers,modules,depth-1);
  callers.pushBack(_owner);
  modules.push_back(_module->getCompleteName());
  }
}

// inline bool Context::started(void) const

bool Context::endReached(void) const
{
int counter=_counter;
int offset;
SmaCode * code;
do
  {
  code=(*_code)[counter];
  if(code->isReturn()) return(true);
  offset=code->jumpsTo(); // is it the end of a 'if' ? (jump across 'else')
  counter+=offset+1;      // the jump code is added
  } while(offset>0); // no jump has been performed --> leave
return(false);
}

//-------- Comparisons -----------------------------------------------------//

//-------- Input/Output ----------------------------------------------------//
Strings Context::printUser(void) const
{
if(_owner) return(_owner->getName().print());
else return(Strings("NONE"));
}

Strings Context::printStack(void) const
{
Strings str;
if(_previous) str+=_previous->printStack();
str+="    '"+printUser()+"' in "+_module->getCompleteName();
if(_code&&_counter&&(*_code)[_counter-1]->getNode())
  {
  const SmaNode * node=(*_code)[_counter-1]->getNode();
  str+="\n        [";
  if(node->getFileName().len())
    str+=node->getFileName();
  else str+="file?";
  str+=":";
  if(node->getLineNumber())
    str+=Strings(node->getLineNumber());
  else str+="line?";
  str+="]\n";
  }
else str+="\n        [file?:line?]\n";
return(str);
}

//-------- Modifiers -------------------------------------------------------//
void Context::enterExecution(Execution * execution)
{
_module->reference();
if(_owner) _owner->reference();
_execution=execution;
_counter=0;
_locals=Context::getValueVect();
_locals->resizeIfNeeded(_module->getNbLocals());
Value * value;
int i;                                 // parameters must have been pushed
for(i=0;i<_module->getNbParams();i++)  // from the last to the first
  {                                   
  if(_module->getParam(i)->getMode()==Parameter::REFERENCE)
    {
    _locals->pushBack(_execution->popValue());
    _locals->accessBack()->reference();
    }
  else // COPY
    {
    value=_execution->popValue();
    _locals->pushBack(LocalValue::getLocalValue(_module->getLocal(i)
                                                       ->getType()));
    _locals->accessBack()->affectCast(*value);
    value->relax();
    }
  }
for(;i<_module->getNbLocals();i++)
  _locals->pushBack(LocalValue::getLocalValue(_module->getLocal(i)
                                                     ->getType()));
}

void Context::leaveExecution(void)
{
unsigned int i;
for(i=0;i<(unsigned int)_module->getNbParams();i++)
  {
  if(_module->getParam(i)->getMode()==Parameter::REFERENCE)
    (*_locals)[i]->relax();
  else // COPY
    LocalValue::relaxLocalValue((LocalValue *)(*_locals)[i]);
  }
for(;i<_locals->getSize();i++)
  LocalValue::relaxLocalValue((LocalValue *)(*_locals)[i]);
Context::relaxValueVect(_locals);
if(_owner) _owner->unReference();
_module->unReference();
}

// inline Execution * Context::accessExecution(void)

// inline void Context::eval(void)

// inline void Context::moveCounter(int offset)

// inline void Context::moveCounterToEnd(void)

// inline Context * Context::accessPrevious(void)

// inline Agent * Context::accessOwner(void)

// inline ExecModule * Context::accessModule(void)

// inline ValueVect * Context::accessLocals(void)

// inline Value * Context::accessLocal(int rank)

void Context::returnValue(Value * value)
{
_execution->pushValue(value);
}

void Context::returnInteger(long integer)
{
Value * value=TmpValue::getTmpValue(_module->getResultType());
value->setInteger(integer);
_execution->pushValue(value);
}

void Context::returnReal(double real)
{
Value * value=TmpValue::getTmpValue(_module->getResultType());
value->setReal(real);
_execution->pushValue(value);
}

void Context::returnStrings(const Strings & str)
{
Value * value=TmpValue::getTmpValue(_module->getResultType());
value->setStrings(str);
_execution->pushValue(value);
}

void Context::returnAgent(Agent * agent,Model * model)
{
Value * value=TmpValue::getTmpValue(ValueType(model));
value->setAgent(agent);
_execution->pushValue(value);
}

// inline void Context::setInteger(int rank,long integer)

// inline void Context::setReal(int rank,double real)

// inline void Context::setStrings(int rank,const Strings & string)

// inline void Context::setAgent(int rank,Agent * agent,Model * model)

// inline void Context::setValueVector(int rank,SubValueVect * vect,
//                                     const ValueType & type)

