// @(#) ENIB/LI2 (c) - oRis - Version 2.3
//--------------------------------------------------------------------------//
// file        : execMod.cpp                                                //
// description : source file for the ENIB/LI2 class ExecModule              //
// author(s)   : F. Harrouet                                                //
// copyright   : Laboratoire d'Informatique Industrielle (LI2)              //
//               Ecole Nationale d'Ingenieurs de Brest (ENIB)               //
//               CP 30815 - 29608 Brest Cedex - FRANCE                      //
//               Tel: (33) 02.98.05.66.31 , Fax: (33) 02.98.05.66.29        //
//               e-mail: li2@enib.fr                                        //
//--------------------------------------------------------------------------//
// Copyright (c) 1996-2000                                                  //
// Laboratoire d'Informatique Industrielle (LI2)                            //
// Ecole Nationale d'Ingenieurs de Brest (ENIB)                             //
//                                                                          //
// Permission to use, copy, modify, distribute and sell this software       //
// and its documentation for any purpose is hereby granted without fee,     //
// provided that the above copyright notice appear in all copies and        //
// that both that copyright notice and this permission notice appear        //
// in supporting documentation.                                             //
// ENIB/LI2 makes no representations about the suitability of this          //
// software for any purpose.                                                //
// It is provided "as is" without express or implied warranty.              //
//------------------------------------------------------------------------- //

#include "oRis/Main/execMod.h"
#include "oRis/Main/protoMod.h"
#include "oRis/Main/context.h"
#include "oRis/Main/simul.h"
#include "oRis/Main/errors.h"
#include "oRis/Nodes/smaNd.h"
#include "oRis/Codes/utilCd.h"

//---------------------------------------------------------------------------//
// class ExecModule                                                          //
//---------------------------------------------------------------------------//

//-------- Allocators -------------------------------------------------------//
ExecModule::ExecModule(const Word & name,
                       ExecModule::SpecialType specialType,
                       const ValueType & resultType,
                       Model * model,Agent * owner)
 : BlockAllocatedObject(),
   _specialType(specialType),
   _locals(), _model(model), _owner(owner),
   _name(name), _resultType(resultType), _nbRefs(0), _params(),
   _mustDisappear(false), _isCritical(false),
   _node((SmaNode *)0), _codeVect((CodeVect *)0)
{
switch(_specialType)
  {
  case ExecModule::CONSTRUCTOR:
  case ExecModule::DESTRUCTOR:
  case ExecModule::EXECUTE:
    _isCritical=true;
    break;
  default:
    break;
  }
}

ExecModule::ExecModule(const ExecModule & module)
 : BlockAllocatedObject(),
   _specialType(module._specialType),
   _locals(module._locals), _model(module._model), _owner(module._owner),
   _name(module._name), _resultType(module._resultType), _nbRefs(0),
   _params(module._params), _mustDisappear(module._mustDisappear),
   _isCritical(module._isCritical),
   _node((SmaNode *)0), _codeVect((CodeVect *)0)
{
if(module._node) _node=module._node->duplicate();
if(module._codeVect&&_node) _codeVect=_node->generateCode();
}

ExecModule * ExecModule::duplicate(void) const // virtual
{
return(new ExecModule(*this));
}

ExecModule::~ExecModule(void)
{
if(_codeVect)
  {
  for(int i=_codeVect->getSize();i--;)
    delete (*_codeVect)[i];
  delete _codeVect;
  }
if(_node) delete _node;
}

//-------- Inspectors ------------------------------------------------------//
// inline const Model * ExecModule::getModel(void) const

// inline const Agent * ExecModule::getOwner(void) const

// inline const Word & ExecModule::getName(void) const

Strings ExecModule::getCompleteName(void) const
{
if(_owner)
  return(_owner->getName().print()+"::"+_name.getStrings()+"()");
if(_model)
  return(_model->getName().getStrings()+"::"+_name.getStrings()+"()");
return("::"+_name.getStrings()+"()");
}

// inline const ValueType & ExecModule::getResultType(void) const

// inline ExecModule::ModuleType ExecModule::getType(void) const // virtual

// inline ExecModule::SpecialType ExecModule::getSpecialType(void) const

// inline bool ExecModule::isCritical(void) const

// inline int ExecModule::getNbParams(void) const

// inline const Parameter * ExecModule::getParam(int rank) const

// inline int ExecModule::getNbLocals(void) const

int ExecModule::getLocalRank(const Word & name) const
{
for(int i=_locals.size();i--;)
  if(_locals[i].getName()==name) return(i);
return(-1);
}

// inline const Attribute * ExecModule::getLocal(int rank) const

// inline const AttrVect & ExecModule::getLocals(void) const

bool ExecModule::isCompatible(const ExecModule & module) const
{
if((_name!=module._name)||
   (_resultType!=module._resultType)||
   (_params.size()!=module._params.size())) return(false);
for(int i=_params.size();i--;)
  {
  if(!_params[i].isCompatible(module._params[i])) return(false);
  }
return(true);
}

Error * ExecModule::tryToCallFromNode(const SmaNode * node,
                                      int nodeOffset) const
{
Strings message;
int minNb;
if(node->getNbNodes()-nodeOffset<(int)_params.size())
  minNb=node->getNbNodes()-nodeOffset;
else minNb=_params.size();
ValueType callType;
for(int i=0;i<minNb;i++)
  {
  callType=node->getNode(i+nodeOffset)->getType();
  if(_params[i].getMode()==Parameter::REFERENCE)
    {
    if(!node->getNode(i+nodeOffset)->isLeftValue()||
       ((callType!=_params[i].getType())&&
        (!callType.canCastTo(_params[i].getType())||
         (callType.getSimpleType()!=ValueType::AGENT))))
      {
      message+="\n        Argument '"+
               _params[i].getName().getStrings()+
               "' should be a '"+
               _params[i].getType().print()+
               " &'";
      }
    }
  else // COPY
    {
    if(!callType.canCastTo(_params[i].getType()))
      {
      message+="\n        Argument '"+
               _params[i].getName().getStrings()+
               "' should be a '"+
               _params[i].getType().print()+
               "'";
      }
    }
  }
if(minNb<(int)_params.size())
  {
  for(unsigned int i=minNb;i<_params.size();i++)
    {
    message+="\n        Missing argument '"+
             _params[i].getName().getStrings()+
             "'";
    }
  }
else if(minNb<node->getNbNodes()-nodeOffset)
  {
  message+="\n        Too many arguments";
  }
if(message.len())
  {
  message="In call to '"+
          getCompleteName()+
          "' :"
          +message;
  return(new Error(Error::SEMANTICS,message));
  }
else return((Error *)0);
}

Error * ExecModule::tryToCall(const Execution * exec,int nbArgs) const
{
Strings message;
int minNb;
if(nbArgs<(int)_params.size())
  minNb=nbArgs;
else minNb=_params.size();
ValueType callType;
for(unsigned int i=0;i<_params.size();i++)
  {
  callType=exec->getValue(i)->getType();
  if(_params[i].getMode()==Parameter::REFERENCE)
    {
    if(!exec->getValue(i)->isLeftValue()||
       ((callType!=_params[i].getType())&&
        (!callType.canCastTo(_params[i].getType())||
         (callType.getSimpleType()!=ValueType::AGENT))))
      {
      message+="\n        Argument '"+
               _params[i].getName().getStrings()+
               "' should be a '"+
               _params[i].getType().print()+
               " &'";
      }
    }
  else // COPY
    {
    if(!callType.canCastTo(_params[i].getType()))
      {
      message+="\n        Argument '"+
               _params[i].getName().getStrings()+
               "' should be a '"+
               _params[i].getType().print()+
               "'";
      }
    }
  }
if(minNb<(int)_params.size())
  {
  for(unsigned int i=minNb;i<_params.size();i++)
    {
    message+="\n        Missing argument '"+
             _params[i].getName().getStrings()+
             "'";
    }
  }
else if(minNb<nbArgs)
  {
  message+="\n        Too many arguments";
  }
if(message.len())
  {
  message="In call to '"+
          getCompleteName()+
          "' :"
          +message;
  return(new Error(Error::SEMANTICS,message));
  }
else return((Error *)0);
}

ProtoModule * ExecModule::buildPrototype(void) const
{
ProtoModule * proto=new ProtoModule(_name,_specialType,
                                    _resultType,_model,_owner);
for(unsigned int i=0;i<_params.size();i++)
  proto->setParam(_params[i]);
return(proto);
}

// inline const CodeVect * ExecModule::getCodeVect(void) const

//-------- Comparisons -----------------------------------------------------//

//-------- Input/Output ----------------------------------------------------//
ostream & operator<<(ostream & stream,const ExecModule & execModule)
{
return(stream << execModule.printProto() << endl
              << execModule.print());
}

Strings ExecModule::print(int column) const // virtual
{
Strings str;
int j;
for(j=0;j<column;j++) str+=" ";
str+=getResultType().print()+" ";
if(getOwner()) str+=getOwner()->getName().print()+"::";
else if(getModel()) str+=getModel()->getName().getStrings()+"::";
str+=getName().getStrings()+_printParams()+"\n";
for(j=0;j<column;j++) str+=" ";
str+="{\n"+printBody(column);
for(j=0;j<column;j++) str+=" ";
str+="}\n";
return(str);
}

Strings ExecModule::printBody(int column) const
{
return(printLocals(column)+printCode(column));
}

Strings ExecModule::printLocals(int column) const
{
Strings str;
for(unsigned int i=getNbParams();i<_locals.size();i++)
  {
  for(int j=0;j<column;j++) str+=" ";
  str+=_locals[i].getType().print()+" "+
       _locals[i].getName().getStrings()+";\n";
  }
return(str);
}

Strings ExecModule::printCode(int column) const
{
return(_node->print(column));
}

Strings ExecModule::printProto(int column) const
{
Strings str;
for(int j=0;j<column;j++) str+=" ";
str+=_resultType.print()+" ";
if(_owner) str+=_owner->getName().print()+"::";
else if(_model) str+=_model->getName().getStrings()+"::";
str+=getName().getStrings()+_printParams()+";\n";
return(str);
}

Strings ExecModule::printModelProto(int column) const
{
Strings str;
for(int j=0;j<column;j++) str+=" ";
str+=_resultType.print()+" "+getName().getStrings()+_printParams()+";\n";
return(str);
}

Strings ExecModule::_printParams(void) const
{
Strings str=Strings("(");
if(_params.empty()) str+=Strings("void");
else
  {
  for(unsigned int i=0;i<_params.size();i++)
    {
    if(i) str+=Strings(",");
    str+=_params[i].print();
    }
  }
str+=Strings(")");
return(str);
}

//-------- Modifiers -------------------------------------------------------//
// inline Model * ExecModule::accessModel(void)

// inline Agent * ExecModule::accessOwner(void)

// inline void ExecModule::setOwner(Agent * owner)

// inline bool & ExecModule::mustDisappear(void)

// inline void ExecModule::setCritical(bool isCritical)

// inline void ExecModule::reference(void)

// inline void ExecModule::unReference(void)

void ExecModule::setParam(const Parameter & param)
{
_params.push_back(param);
_locals.push_back(Attribute(param.getName(),param.getType(),
                            (Model *)0,(Agent *)0));
}

// inline void ExecModule::setLocal(const Attribute & attr)

void ExecModule::setNode(SmaNode * node)
{
if(_node) delete _node;
if(_codeVect)
  {
  for(int i=_codeVect->getSize();i--;)
    delete (*_codeVect)[i];
  delete _codeVect;
  }
_node=node;
_codeVect=_node->generateVoidCode();
_codeVect->pushBack(new ReturnCode(_node,0,0));
}

