// @(#) ENIB/LI2 (c) - oRis - Version 2.3
//--------------------------------------------------------------------------//
// file        : model.cpp                                                  //
// description : source file for the ENIB/LI2 class Model                   //
// author(s)   : F. Harrouet                                                //
// copyright   : Laboratoire d'Informatique Industrielle (LI2)              //
//               Ecole Nationale d'Ingenieurs de Brest (ENIB)               //
//               CP 30815 - 29608 Brest Cedex - FRANCE                      //
//               Tel: (33) 02.98.05.66.31 , Fax: (33) 02.98.05.66.29        //
//               e-mail: li2@enib.fr                                        //
//--------------------------------------------------------------------------//
// Copyright (c) 1996-2000                                                  //
// Laboratoire d'Informatique Industrielle (LI2)                            //
// Ecole Nationale d'Ingenieurs de Brest (ENIB)                             //
//                                                                          //
// Permission to use, copy, modify, distribute and sell this software       //
// and its documentation for any purpose is hereby granted without fee,     //
// provided that the above copyright notice appear in all copies and        //
// that both that copyright notice and this permission notice appear        //
// in supporting documentation.                                             //
// ENIB/LI2 makes no representations about the suitability of this          //
// software for any purpose.                                                //
// It is provided "as is" without express or implied warranty.              //
//------------------------------------------------------------------------- //

#include "oRis/Main/model.h"
#include "oRis/Main/simul.h"
#include "oRis/Main/tmpVal.h"
#include "oRis/Main/errors.h"
#include "oRis/Main/execMod.h"

//---------------------------------------------------------------------------//
// class Model                                                               //
//---------------------------------------------------------------------------//

Model * Model::_baseModel=(Model *)0; // static
Word Model::_notifyAncestorWord(Word::create("notifyAncestor"));     // static
Word Model::_notifyAttributeWord(Word::create("notifyAtribute"));    // static
Word Model::_notifyModuleWord(Word::create("notifyModule"));         // static
Word Model::_classDescriptionWord(Word::create("ClassDescription")); // static
Word Model::_classNameWord(Word::create("className"));               // static

//-------- Static methods --------------------------------------------------//
// inline Model * Model::getBaseModel(void) // static

//-------- Allocators -------------------------------------------------------//
Model::Model(const Word & name)
 : BlockAllocatedObject(),
   _name(name), _rank(getSimul()->getNbModels()), _native(false), _instNum(0),
   _parents(), _ancestors(), _children(), _descendants(),
   _agentMap(), _attrMap(), _moduleMap(), _mainModule((ExecModule *)0),
   _newModule((ExecModule *)0), _deleteModule((ExecModule *)0),
   _broadcastList()
{
if(_rank) setParent(_baseModel,false); // if not base model inherit from
else _baseModel=this;                  // else become the base model
}

Model::~Model(void)
{
ExecModMap::iterator it=_moduleMap.begin();
for(;!(it==_moduleMap.end());it++)
  (*it).second->unReference();
if(_newModule) _newModule->unReference();
if(_deleteModule) _deleteModule->unReference();
while(_broadcastList.size())
  {
  delete _broadcastList.back();
  _broadcastList.pop_back();
  }
}

//-------- Inspectors ------------------------------------------------------//
// inline const Word & Model::getName(void) const

// inline int Model::getRank(void) const

// inline bool Model::isNative(void) const

// inline const ModelVect & Model::getParents(void) const

bool Model::hasParent(const Model * model) const
{
for(int i=_parents.getSize();i--;)
  if(_parents[i]==model) return(true);
return(false);
}

bool Model::hasParent(const Word & name) const
{
return(hasParent(getSimul()->getModel(name)));
}

// inline const ModelVect & Model::getAncestors(void) const

bool Model::hasAncestor(const Model * model) const
{
for(int i=_ancestors.getSize();i--;)
  if(_ancestors[i]==model) return(true);
return(false);
}

bool Model::hasAncestor(const Word & name) const
{
return(hasAncestor(getSimul()->getModel(name)));
}

// inline const ModelVect & Model::getChildren(void) const

bool Model::hasChildren(const Model * model) const
{
for(int i=_children.getSize();i--;)
  if(_children[i]==model) return(true);
return(false);
}

bool Model::hasChildren(const Word & name) const
{
return(hasChildren(getSimul()->getModel(name)));
}

// inline const ModelVect & Model::getDescendants(void) const

bool Model::hasDescendant(const Model * model) const
{
for(int i=_descendants.getSize();i--;)
  if(_descendants[i]==model) return(true);
return(false);
}

bool Model::hasDescendant(const Word & name) const
{
return(hasDescendant(getSimul()->getModel(name)));
}

// inline bool Model::isA(const Model * model) const

bool Model::isA(const Word & name) const
{
return(isA(getSimul()->getModel(name)));
}

Error * Model::tryToSetParent(const Model * parent) const
{
if(isA(parent)) return((Error *)0);
if(parent->isA(this))
  {
  Strings message="Class '"+
                  parent->getName().getStrings()+
                  "' already inherits from class '"+
                  _name.getStrings()+
                  "'";
  return(new Error(Error::SEMANTICS,message));
  }
if(parent->_native)
  {
  if(_native)
    {
    Strings message="Native class '"+
                    _name.getStrings()+
                    "' can't inherit from native class '"+
                    parent->getName().getStrings()+
                    "'";
    return(new Error(Error::SEMANTICS,message));
    }
  for(int i=_descendants.getSize();i--;)
    {
    if(_descendants[i]->_native)
      {
      Strings message="Native subclass '"+
                      _descendants[i]->getName().getStrings()+
                      "' of class '"+
                      _name.getStrings()+
                      "' can't inherit from native class '"+
                      parent->getName().getStrings()+
                      "'";
      return(new Error(Error::SEMANTICS,message));
      }
    }
  }
if((_newModule&&_newModule->getType()!=ExecModule::PROTOTYPE)||
   (_deleteModule&&_deleteModule->getType()!=ExecModule::PROTOTYPE))
  {
  Strings message="Class '"+
                  _name.getStrings()+
                  "' can't inherit from class '"+
                  parent->getName().getStrings()+
                  "' ";
  if(!_newModule||
     _newModule->getType()==ExecModule::PROTOTYPE)
    message+="(destructor is already defined)";
  else if(!_deleteModule||
          _deleteModule->getType()==ExecModule::PROTOTYPE)
    message+="(constructor is already defined)";
  else
    message+="(constructor and destructor are already defined)";
  return(new Error(Error::SEMANTICS,message));
  }
Error * error=(Error *)0;
AttrMap::const_iterator itA=parent->_attrMap.begin();
for(;!error&&!(itA==parent->_attrMap.end());itA++)
  error=tryToSetAttribute((*itA).second);
ExecModMap::const_iterator itM=parent->_moduleMap.begin();
for(;!error&&!(itM==parent->_moduleMap.end());itM++)
  error=tryToSetModule((*itM).second);
return(error);
}

// inline long Model::getInstNum(void) const

// inline int Model::getNbAgents(void) const

// inline const AgentMap & Model::getAgentMap(void) const

// inline int Model::getNbAttributes(void) const

// inline const AttrMap & Model::getAttrMap(void) const

const Attribute * Model::getAttribute(const Word & name) const
{
AttrMap::const_iterator it=_attrMap.find(name);
if(it==_attrMap.end()) return((const Attribute *)0);
return(&(*it).second);
}

Error * Model::tryToSetAttribute(const Attribute & attr) const
{
Error * error=_tryToSetAttribute(attr);                   // try in model ...
for(unsigned int i=0;
    !error&&(i<_descendants.getSize());i++)               // ... descendants
  error=_descendants[i]->_tryToSetAttribute(attr);
AgentMap::const_iterator it=_agentMap.begin();
for(;!error&&!(it==_agentMap.end());it++)                 // ... agents
  error=(*it).second->tryToSetAttribute(attr);
return(error);
}

// inline int Model::getNbModules(void) const

// inline const ExecModMap & Model::getModuleMap(void) const

// inline const ExecModule * Model::getMainModule(void) const

// inline const ExecModule * Model::getNewModule(void) const

// inline const ExecModule * Model::getDeleteModule(void) const

const ExecModule * Model::getModule(const Word & name) const
{
ExecModMap::const_iterator it=_moduleMap.find(name);
if(it==_moduleMap.end())
  {
  if(_newModule&&_newModule->getName()==name) return(_newModule);
  if(_deleteModule&&_deleteModule->getName()==name) return(_deleteModule);
  return((const ExecModule *)0);
  }
return((*it).second);
}

Error * Model::tryToSetModule(const ExecModule * newModule) const
{
Error * error=_tryToSetModule(newModule);                 // try in model ...
for(unsigned int i=0;
    !error&&(i<_descendants.getSize());i++)               // ... descendants
  error=_descendants[i]->_tryToSetModule(newModule);
AgentMap::const_iterator it=_agentMap.begin();
for(;!error&&!(it==_agentMap.end());it++)                 // ... agents
  error=(*it).second->tryToSetModule(newModule);
return(error);
}

//-------- Comparisons -----------------------------------------------------//
// inline bool Model::operator==(const Model & model) const

// inline bool Model::operator!=(const Model & model) const

// inline bool Model::operator<(const Model & model) const

// inline bool Model::operator<=(const Model & model) const

// inline bool Model::operator>=(const Model & model) const

// inline bool Model::operator>(const Model & model) const

//-------- Input/Output ----------------------------------------------------//
ostream & operator<<(ostream & stream, const Model & model)
{
return(stream << model.printDeclaration(0) << endl
              << model.printModules(0)     << endl);
}

Strings Model::printDeclaration(int column) const
{
Strings str;
int j;
for(j=0;j<column;j++) str+=" ";
str+=Strings("class ")+_name.getStrings();
for(unsigned int i=1;i<_parents.getSize();i++)
  {
  if(i>1) str+=",";
  else str+=" : ";
  str+=_parents[i]->getName().getStrings();
  }
str+="\n";
for(j=0;j<column;j++) str+=" ";
str+="{\n";
AttrMap::const_iterator itA=_attrMap.begin();
for(;!(itA==_attrMap.end());itA++)
  {
  for(j=0;j<column;j++) str+=" ";
  str+=(*itA).second.getType().print()+" "+
       (*itA).second.getName().getStrings()+";\n";
  }
if(_newModule) str+=_newModule->printModelProto(column);
if(_deleteModule) str+=_deleteModule->printModelProto(column);
ExecModMap::const_iterator itM=_moduleMap.begin();
for(;!(itM==_moduleMap.end());itM++)
  {
  if((*itM).second->getModel()==this)
    str+=(*itM).second->printModelProto(column);
  }
for(j=0;j<column;j++) str+=" ";
str+="};\n";
return(str);
}

Strings Model::printModules(int column) const
{
Strings str;
if(_newModule) str+=_newModule->print(column)+"\n";
if(_deleteModule) str+=_deleteModule->print(column)+"\n";
ExecModMap::const_iterator it=_moduleMap.begin();
for(;!(it==_moduleMap.end());it++)
  {
  if(((*it).second->getModel()==this)&&
     ((*it).second->getType()!=ExecModule::PROTOTYPE))
    str+=(*it).second->print(column)+"\n";
  }
return(str);
}

//-------- Modifiers -------------------------------------------------------//
// inline void Model::setNative(void)

// inline ModelVect & Model::accessParents(void)

// inline ModelVect & Model::accessAncestors(void)

// inline ModelVect & Model::accessChildren(void)

// inline ModelVect & Model::accessDescendants(void)

void Model::setParent(Model * parent,bool notify)
{
if(_setParent(this,parent))                     // store in model ...
  {
  if(parent->_native) setNative();
  int i;
  for(i=_descendants.getSize();i--;)            // ... descendants
    _descendants[i]->_setParent(this,parent);
  AttrMap::iterator itA=parent->_attrMap.begin();
  for(;!(itA==parent->_attrMap.end());itA++)    // store attributes
    setAttribute((*itA).second,notify);
  ExecModMap::iterator itM=parent->_moduleMap.begin();
  for(;!(itM==parent->_moduleMap.end());itM++)  // store modules
    setModule((*itM).second,notify);
  if(notify)
    {
    for(i=_descendants.getSize();i--;)
      _descendants[i]->_notifyChange(_notifyAncestorWord,parent->_name);
    _notifyChange(_notifyAncestorWord,parent->_name);
    }
  }
}

// inline AgentMap & Model::accessAgentMap(void)

void Model::setAttribute(const Attribute & attr,bool notify)
{
if(_setAttribute(attr))                               // store in model ...
  {
  unsigned int i;
  for(i=0;i<_descendants.getSize();i++)              // ... descendants
    _descendants[i]->_setAttribute(attr);
  AgentMap::iterator it=_agentMap.begin();
  for(;!(it==_agentMap.end());it++)                  // ... agents
    (*it).second->setAttribute(attr,notify);
  if(notify)
    {
    for(i=_descendants.getSize();i--;)
      _descendants[i]->_notifyChange(_notifyAttributeWord,attr.getName());
    _notifyChange(_notifyAttributeWord,attr.getName());
    }
  }
}

// inline AttrMap & Model::accessAttrMap(void)

Attribute * Model::accessAttribute(const Word & name)
{
AttrMap::iterator it=_attrMap.find(name);
if(it==_attrMap.end()) return((Attribute *)0);
return(&(*it).second);
}

void Model::setModule(ExecModule * newModule,bool notify)
{
if(_setModule(newModule))                            // store in model ...
  {
  unsigned int i;
  for(i=0;i<_descendants.getSize();i++)              // ... descendants
    _descendants[i]->_setModule(newModule);
  AgentMap::iterator it=_agentMap.begin();
  for(;!(it==_agentMap.end());it++)                  // ... agents
    (*it).second->setModule(newModule,notify);
  if(notify)
    {
    for(i=_descendants.getSize();i--;)
      _descendants[i]->_notifyChange(_notifyModuleWord,newModule->getName());
    _notifyChange(_notifyModuleWord,newModule->getName());
    }
  }
}
  
// inline ExecModMap & Model::accessModuleMap(void)

// inline ExecModule * Model::accessMainModule(void)

// inline ExecModule * Model::accessNewModule(void)

// inline ExecModule * Model::accessDeleteModule(void)

ExecModule * Model::accessModule(const Word & name)
{
ExecModMap::iterator it=_moduleMap.find(name);
if(it==_moduleMap.end())
  {
  if(_newModule&&_newModule->getName()==name) return(_newModule);
  if(_deleteModule&&_deleteModule->getName()==name) return(_deleteModule);
  return((ExecModule *)0);
  }
return((*it).second);
}

//-------- Agent Creation/Deletion -----------------------------------------//
Agent * Model::createAgent(long instNum)
{
if(!instNum) instNum=++_instNum;
else if(instNum>_instNum) _instNum=instNum;
Agent * newAgent=getSimul()->getNewAgent(this,instNum); // link oRis <---> C++
_agentMap.insert(AgentMap::value_type(newAgent->getName(),newAgent));
for(int i=_ancestors.getSize();i--;)
  _ancestors[i]->_agentMap.insert(AgentMap::value_type(
                                  newAgent->getName(),newAgent));
return(newAgent);
}

void Model::removeAgent(Agent * agent)
{
_agentMap.erase(agent->getName());
for(int i=_ancestors.getSize();i--;)
  _ancestors[i]->_agentMap.erase(agent->getName());
}

//-------- Broadcast -------------------------------------------------------//
void Model::getSensitivity(Agent * sensitive,
                           ModelVect & messages,StringsVect & methods)
{
BroadcastList::iterator it=_findBroadcastInfo(sensitive);
if((it!=_broadcastList.end())&&
   ((*it)->_msgType==this))
  {
  messages.pushBack((*it)->_msgType);
  methods.push_back((*it)->_method.getStrings());
  }
for(unsigned int i=0;i<_descendants.getSize();i++)
  {
  it=_descendants[i]->_findBroadcastInfo(sensitive);
  if((it!=_descendants[i]->_broadcastList.end())&&
     ((*it)->_msgType==_descendants[i]))
    {
    messages.pushBack((*it)->_msgType);
    methods.push_back((*it)->_method.getStrings());
    }
  }
}

void Model::setSensitivity(Agent * sensitive,const Word & method)
{
BroadcastList::iterator it=_findBroadcastInfo(sensitive);
if(it!=_broadcastList.end())
  {
  (*it)->_msgType=this;
  (*it)->_method=method;
  }
else _broadcastList.push_back(new BroadcastInfo(this,sensitive,method));
for(int i=_descendants.getSize();i--;)
  {
  it=_descendants[i]->_findBroadcastInfo(sensitive);
  if(it!=_descendants[i]->_broadcastList.end())
    {
    if(isA((*it)->_msgType))
      {
      (*it)->_msgType=this;
      (*it)->_method=method;
      }
    }
  else
    _descendants[i]->_broadcastList.push_back(
                     new BroadcastInfo(this,sensitive,method));
  }
}

void Model::cancelSensitivity(Agent * sensitive)
{
bool found;
int i;
BroadcastList::iterator upper=(BroadcastList::iterator)0;
BroadcastList::iterator it=_findBroadcastInfo(sensitive);
if((it!=_broadcastList.end())&&
   ((*it)->_msgType==this))
  {
  found=false;
  for(i=_parents.getSize();i--;)
    {
    upper=_parents[i]->_findBroadcastInfo(sensitive);
    if(upper!=_parents[i]->_broadcastList.end())
      {
      found=true;
      break;
      }
    }
  if(found)
    {
    (*it)->_msgType=(*upper)->_msgType;
    (*it)->_method=(*upper)->_method;
    }
  else
    {
    delete *it;
    _broadcastList.erase(it);
    }
  for(i=_descendants.getSize();i--;)
    {
    it=_descendants[i]->_findBroadcastInfo(sensitive);
    if((it!=_descendants[i]->_broadcastList.end())&&
       ((*it)->_msgType==this))
      {
      if(found)
        {
        (*it)->_msgType=(*upper)->_msgType;
        (*it)->_method=(*upper)->_method;
        }
      else
        {
        delete *it;
        _descendants[i]->_broadcastList.erase(it);
        }
      }
    }
  }
}

void Model::removeSensitive(Agent * sensitive)
{
BroadcastList::iterator it=_findBroadcastInfo(sensitive);
if(it!=_broadcastList.end())
  {
  delete *it;
  _broadcastList.erase(it);
  }
for(int i=_descendants.getSize();i--;)
  {
  it=_descendants[i]->_findBroadcastInfo(sensitive);
  if(it!=_descendants[i]->_broadcastList.end())
    {
    delete *it;
    _descendants[i]->_broadcastList.erase(it);
    }
  }
}

void Model::broadcast(Agent * message)
{
Execution * initExec=Execution::getExecution(message,ExecModule::NORMAL);
Value * value;
ExecModule * module;
BroadcastInfo * info;
bool useFront;
BroadcastList::iterator it;
BroadcastList tmpList(_broadcastList);
while(tmpList.size())
  {
  if((getSimul()->getPickingMode()==Simulator::SERIAL)||
     Simulator::integerRand(2))
    {
    useFront=true;
    info=tmpList.front();
    }
  else
    {
    useFront=false;
    info=tmpList.back();
    }
  for(it=_broadcastList.begin();it!=_broadcastList.end();it++)
    {
    if((*it)==info)
      {
      module=info->_sensitive->accessModule(info->_method);
      if(module->getType()!=ExecModule::PROTOTYPE)
        {
        value=TmpValue::getTmpValue(ValueType(message->accessModel()));
        value->setAgent(message);
        initExec->pushValue(value);
        getSimul()->executeModule(module,info->_sensitive,initExec);
        }
      break;
      }
    }
  if(useFront) tmpList.pop_front();
  else tmpList.pop_back();
  }
Execution::relaxExecution(initExec);
}

//-------- Private Methods -------------------------------------------------//
Error * Model::_tryToSetAttribute(const Attribute & attr) const
{
const Attribute * oldAttr=getAttribute(attr.getName());
if(oldAttr&&oldAttr->getType()!=attr.getType())
  {
  Strings message="Attribute '"+
                  oldAttr->getName().getStrings()+
                  "' has type '"+
                  oldAttr->getType().print()+
                  "' in '"+
                  _name.getStrings()+
                  "'";
  return(new Error(Error::SEMANTICS,message));
  }
return((Error *)0);
}

Error * Model::_tryToSetModule(const ExecModule * newModule) const
{
switch(newModule->getSpecialType())
  {
  case ExecModule::NORMAL:
    {
    const ExecModule * oldModule=getModule(newModule->getName());
    if(oldModule&&!oldModule->isCompatible(*newModule))
      {
      Strings message="Module '"+
                      newModule->getCompleteName()+
                      "' is incompatible with '"+
                      oldModule->getCompleteName()+
                      "'";
      return(new Error(Error::SEMANTICS,message));
      }
    }
    break;
  case ExecModule::BEHAVIOR:
    if((newModule->getResultType()!=ValueType())||
       newModule->getNbParams())
      {
      Strings message="Prototype of module '"+
                      newModule->getCompleteName()+
                      "' should be 'void "+
                      newModule->getName().getStrings()+
                      "(void)'";
      return(new Error(Error::SEMANTICS,message));
      }
    if(_mainModule&&!_mainModule->isCompatible(*newModule))
      {
      Strings message="Module '"+
                      newModule->getCompleteName()+
                      "' is incompatible with '"+
                      _mainModule->getCompleteName()+
                      "'";
      return(new Error(Error::SEMANTICS,message));
      }
    break;
  case ExecModule::CONSTRUCTOR:
    if(newModule->getModel()==this) // ignore ancestors
      {
      if(newModule->getResultType()!=ValueType())
        {
        Strings message="Prototype of module '"+
                        newModule->getCompleteName()+
                        "' should be 'void "+
                        newModule->getName().getStrings()+
                        "(...)'";
        return(new Error(Error::SEMANTICS,message));
        }
      if(_newModule&&!_newModule->isCompatible(*newModule))
        {
        Strings message="Module '"+
                        newModule->getCompleteName()+
                        "' is incompatible with '"+
                        _newModule->getCompleteName()+
                        "'";
        return(new Error(Error::SEMANTICS,message));
        }
      }
    break;
  case ExecModule::DESTRUCTOR:
    if(newModule->getModel()==this) // ignore ancestors
      {
      if((newModule->getResultType()!=ValueType())||
         newModule->getNbParams())
        {
        Strings message="Prototype of module '"+
                        newModule->getCompleteName()+
                        "' should be 'void "+
                        newModule->getName().getStrings()+
                        "(void)'";
        return(new Error(Error::SEMANTICS,message));
        }
      if(_deleteModule&&!_deleteModule->isCompatible(*newModule))
        {
        Strings message="Module '"+
                        newModule->getCompleteName()+
                        "' is incompatible with '"+
                        _deleteModule->getCompleteName()+
                        "'";
        return(new Error(Error::SEMANTICS,message));
        }
      }
    break;
  default:
    break;
  }
return((Error *)0);
}

bool Model::_setParent(Model * level,Model * parent)
{
if(isA(parent)) return(false);
unsigned int pos;
for(pos=0;(pos<_ancestors.getSize())&&
          (_ancestors[pos]!=level);pos++);   // insert here
_ancestors.insert(pos,parent);               // becomes an ancestor
parent->_descendants.pushBack(this);
if(level==this)                              // becomes a parent
  {
  if((_parents.getSize()==1)&&
     (_parents[0]==Model::_baseModel)) // first inheritance ...
    {
    for(unsigned int p=0;p<_parents[0]->_children.getSize();p++)
      {
      if(_parents[0]->_children[p]==this)
        {                 // not a child of 'Object' anymore
        _parents[0]->_children.erase(p);
        break;
        }
      }
    _parents[0]=parent; // 'Object' is not a parent anymore
    }
  else _parents.pushBack(parent);
  parent->_children.pushBack(this);
  }
for(int i=parent->_ancestors.getSize();i--;)         // get new ancestors
  if(!hasAncestor(parent->_ancestors[i]))
    {
    _ancestors.insert(pos,parent->_ancestors[i]);
    parent->_ancestors[i]->_descendants.pushBack(this);
    }
BroadcastList::iterator itS1,itS2;                // get new sensitives
for(itS1=parent->_broadcastList.begin();
    itS1!=parent->_broadcastList.end();itS1++)
  {
  itS2=_findBroadcastInfo((*itS1)->_sensitive);
  if(itS2!=_broadcastList.end())
    {
    if((*itS2)->_msgType!=this) // local is not more recent
      {
      (*itS2)->_msgType=(*itS1)->_msgType;
      (*itS2)->_method=(*itS1)->_method;
      }
    }
  else _broadcastList.push_back(new BroadcastInfo(**itS1));
  }
return(true);
}

bool Model::_setAttribute(const Attribute & attr)
{
if(getAttribute(attr.getName())) return(false);
else // not found --> insert
  {
  _attrMap.insert(AttrMap::value_type(attr.getName(),attr));
  return(true);
  }
}

bool Model::_setModule(ExecModule * newModule)
{
bool used=false;
switch(newModule->getSpecialType())
  {
  case ExecModule::NORMAL:
  case ExecModule::BEHAVIOR:
    {
    ExecModMap::iterator it;
    it=_moduleMap.find(newModule->getName());
    if(it==_moduleMap.end())                 // not found --> insert
      {
      _moduleMap.insert(ExecModMap::value_type(newModule->getName(),
                                               newModule));
      used=true;
      newModule->reference();
      if(newModule->getSpecialType()==ExecModule::BEHAVIOR)
        _mainModule=newModule;
      }
    else                                     // found --> replace if needed
      {
      ExecModule * oldModule=(*it).second;
      if((oldModule!=newModule)&&                           // not the same
         newModule->getModel()->isA(oldModule->getModel())) // and more recent
        {
        oldModule->unReference();
        (*it).second=newModule;
        used=true;
        newModule->reference();
        if(newModule->getSpecialType()==ExecModule::BEHAVIOR)
          _mainModule=newModule;
        }
      }
    }
    break;
  case ExecModule::CONSTRUCTOR:
    if((newModule->getModel()==this)&& // ignore ancestors
       (newModule!=_newModule))        // not the same
      {
      if(_newModule) _newModule->unReference();
      _newModule=newModule;
      used=true;
      _newModule->reference();
      }
    break;
  case ExecModule::DESTRUCTOR:
    if((newModule->getModel()==this)&& // ignore ancestors
       (newModule!=_deleteModule))     // not the same
      {
      if(_deleteModule) _deleteModule->unReference();
      _deleteModule=newModule;
      used=true;
      _deleteModule->reference();
      }
    break;
  default:
    break;
  }
return(used);
}

void Model::_notifyChange(const Word & method,const Word & param)
{
Model * model=getSimul()->accessModel(_classDescriptionWord);
if(model)
  {
  Agent * descr;
  Execution * initExec;
  ExecModule * module;
  Value * value;
  StlVector<AgentName> agtVect;
  AgentMap::const_iterator it=model->getAgentMap().begin();
  for(;it!=model->getAgentMap().end();it++)
    agtVect.push_back((*it).first);
  while(agtVect.size())
    {
    descr=getSimul()->accessAgent(agtVect.back());
    agtVect.pop_back();
    if(descr&&
       (descr->getAttribute(_classNameWord)->getValue()->getStrings()
        ==getName().getStrings()))
      {
      module=descr->accessModule(method);
      if(module->getType()!=ExecModule::PROTOTYPE)
        {
        initExec=Execution::getExecution(descr,ExecModule::NORMAL);
        value=TmpValue::getTmpValue(ValueType(ValueType::STRING));
        value->setStrings(param.getStrings());
        initExec->pushValue(value);
        getSimul()->executeModule(module,descr,initExec);
        Execution::relaxExecution(initExec);
        }
      }
    }
  }
}

BroadcastList::iterator Model::_findBroadcastInfo(Agent * sensitive)
{
BroadcastList::iterator it=_broadcastList.begin();
for(;it!=_broadcastList.end();it++)
  {
  if((*it)->_sensitive==sensitive) break;
  }
return(it);
}

