//-------------------------------------------------------------------------
//
// File    	:  fuzzySet.h
// Description  :  ensemble flou
//
// Author(s)   	: P. Reignier
// Date         : 09/03/97
// Copyright	: Laboratoire d'Informatique Industrielle (LI2)
//		  Ecole Nationale d'Ingenieurs de Brest (ENIB)
//		  CP 15 - 29608 Brest Cedex - FRANCE
//		  Tel: (33) 98.05.66.31, Fax: (33) 98.05.66.29
//		  e-mail: li2@li2.enib.fr
//-------------------------------------------------------------------------
// Copyright (c) 1996
// Laboratoire d'Informatique Industrielle (LI2)
// Ecole Nationale d'Ingenieurs de Brest (ENIB)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.
// ENIB/LI2 makes no representations about the suitability of this
// software for any purpose.
// It is provided "as is" without express or implied warranty.
//-------------------------------------------------------------------------


#ifndef _FUZZY_SET
#define _FUZZY_SET

#ifndef __MSDOS__
// #include <stl_config.h>
#endif

#include "oRis/Main/agent.h"
#include "Utility/stlUtil.h"
#include "Utility/stlVect.h"
#include "Utility/stlMap.h"
#include "point2D.h"

class Model ;

//----------------------------------------------------------------------//

class FuzzySet : public BlockAllocatedObject
{
public:
  typedef StlMap< double, StlPair<double,double>, StlLess<double> > Fusion ;
  enum Type {CRISP, FUZZY} ;

  // Constructors and destructor
  virtual ~FuzzySet (void) ;

  // native constructors

  FuzzySet (double value = 0.0) ;          // init with a crisp value
  FuzzySet (const StlVector<Point2D>&) ;   // init with a fuzzy value
  FuzzySet (const FuzzySet&) ;             // init with a copy

  // Operators

  FuzzySet& operator= (const FuzzySet&) ; // affectation
  friend ostream& operator<< (ostream&, const FuzzySet&) ;
  // injection operator

  // Modifiers
  
  void setValue(double) ;            // set the crisp value of the set 
  void setValue(const StlVector<Point2D>&) ;
  // set the fuzzy value of the set 
  void saturate(double a) ;  // saturate the membership function to a
  void product(double a) ;   // multiply the membership function by a

  void combineMax (const FuzzySet& s) ;  // combine the set with s
  void combineMin (const FuzzySet& s) ;  // combine the set with s
  void combineSum (const FuzzySet& s) ;  // combine the set with s

  // Observers 

  double mu (double) const ;     
  // membership function associated to the set 
  double integral(void) const ;     
  // integrates the membership function         
  double xIntegral(void) const ;
  // integrates x * membership function
  bool maxMu(StlVector<Point2D>&) const ;
  // list of points where mu is maximum
  double maxMin(const FuzzySet& set) const ;
  // returns the max of the min between this and the set

  double getCrispValue(void) const ;    
  // returns the crisp value of the set
  const StlVector<Point2D>& getFuzzyValue(void) const ;
  // returns the fuzzy value of the set
  const Point2D& getValue(int)  const ;
  // get the nieme point describing the membership function
  int getSize(void) const ;
  inline Type getType() const {return (_type) ;} ;
  // returns the type of the set 

  void simplify(void) ; // suppress all multiple successive points 
  // on the same level
  void normalize(void); // normalize the fuzzy set

  // linguistic modifiers
  FuzzySet fNot(void) const ;
  FuzzySet fMoreOrLess(void) const ;
  FuzzySet fVery(void) const ;

protected:
  // Utilities
  void _destroy(void) ;
  void _copy(const FuzzySet&) ;

  double _value;// value of the associated crisp number if the set is crisp 
  Type _type ;  // type of the fuzzy set
  StlVector<Point2D> _listePoints ;  // list of edges on the membership
  // function.


  // fuzzy functions utilities

  double _feval(const StlVector<Point2D>& liste, double x) const ;

  void _merge(const StlVector<Point2D>& vec1, 
	      const StlVector<Point2D>& vec2,
	      Fusion& res) const ;
  void _intersect(Fusion& fonction) const ; // complete the given two functions
  // with their corresponding intersection points
  void _razListePoints (void) 
    {
      while(_listePoints.size() != 0)
	_listePoints.pop_back() ;
    }
} ;

#endif

  
  
