//-------------------------------------------------------------------------
//
// File    	:  fuzzyController.cpp
// Description  :  fuzzy Controller
//
// Author(s)   	: P. Reignier
// Date         : 09/03/97
// Copyright	: Laboratoire d'Informatique Industrielle (LI2)
//		  Ecole Nationale d'Ingenieurs de Brest (ENIB)
//		  CP 15 - 29608 Brest Cedex - FRANCE
//		  Tel: (33) 98.05.66.31, Fax: (33) 98.05.66.29
//		  e-mail: li2@li2.enib.fr
//-------------------------------------------------------------------------
// Copyright (c) 1996
// Laboratoire d'Informatique Industrielle (LI2)
// Ecole Nationale d'Ingenieurs de Brest (ENIB)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.
// ENIB/LI2 makes no representations about the suitability of this
// software for any purpose.
// It is provided "as is" without express or implied warranty.
//-------------------------------------------------------------------------

#ifndef __MSDOS__
#include <sstream>
#else
#include <strstrea.h>
#endif
#include <fstream>
#include <string.h>

#include "oRis/Main/simul.h"
#include "oRis/Main/errors.h"
#include "oRis/Main/agent.h"
#include "oRis/Main/value.h"
#include "oRis/Main/model.h"
#include "oRis/Main/context.h"
#include "fuzzyController.h"
#include "fuzzyFNod.h" 
#include "Utility/stlSet.h"

class FuzzyParser ;
#include "fuzzyScan.h"
#include "fuzzyParse.h"

class FuzzyParserLoad ;
#include "fuzzyScanLoad.h"
#include "fuzzyParseLoad.h"

//----------------------------------------------------------------------//
// Constructors and destructors
//----------------------------------------------------------------------//

FuzzyController::FuzzyController(Model* model, long instNum) : 
  Agent(model, instNum) 
{
  _rules = accessAttribute(Word::create("_rules"))->accessValue() ;
  _usedVariables = accessAttribute(Word::create("_usedVariables"))
                   ->accessValue() ;
  _destroyed = false ;
}

//----------------------------------------------------------------------//

FuzzyController::~FuzzyController(void)
{
  if (!_destroyed)
    destroy() ;
}

//----------------------------------------------------------------------//
// Utilities
//----------------------------------------------------------------------//

void FuzzyController::destroy(void) 
{
  _destroyed = true ;
}

//----------------------------------------------------------------------//

void FuzzyController::addUsedVariable(Agent* var)
{
  _usedVariablesSet.insert(var) ; 
}

//----------------------------------------------------------------------//

void FuzzyController::compile(char* rules, bool atomic)
{
  _usedVariablesSet.clear() ;

  FuzzyFNod* root = (FuzzyFNod*) 0;
  istringstream input(rules);

  FuzzyParser parse(&input, &root) ;
  parse.yyparse() ;

  FNodData data ;
  data.clearoRisProg() ;

  data.resetTmpVars() ;
  data.setFuzzyController(this) ;

  if (root)
    root->generate(data) ;

  SubValueVect* usedVariablesVect = SubValue::getSubValueVect() ;
  
  StlSet < Agent * > ::iterator i ;
  for (i=_usedVariablesSet.begin(); i != _usedVariablesSet.end(); i++)
  {
     usedVariablesVect->pushBack(SubValue::getSubValue(ValueType(ValueType((*i)->accessModel())),
                                                      _usedVariables));
     usedVariablesVect->accessBack()->setAgent((*i));
  }
  _usedVariables->setValueVector(usedVariablesVect) ;
  
  if (!data.isError())
    {
      Strings method ;

	  // we redefine the FuzzyController::eval()
	  // even if we are in a derivate class 
      method = Strings("void ")+getName().print() + 
    	Strings("::_eval(void) {") ;
      if (atomic)
	      method += Strings("execute {") ;
      method += data.getoRisProgram() ;
      if (atomic)
      	method += Strings("return;\n}") ;

      method += Strings("}") ;
      getSimul()->parseStrings(method) ;
    }

}

//----------------------------------------------------------------------//

void FuzzyController::addRule(const Strings rule)
{
  _rules->pushBack(SubValue::getSubValue(ValueType(ValueType::STRING),
					 _rules)) ;
  _rules->accessBack()->setStrings(rule) ;
}

//----------------------------------------------------------------------//

void FuzzyController::addRulesFromFile(const Strings& filename)
{
  ifstream in((const char *) filename) ;
  if (in.fail())
    {
      Strings message = filename + " : file not found " ;
      getSimul()->manageError(new Error(Error::RUNTIME, message)); 
      return ;
    }
  FuzzyParserLoad parser(&in, this) ;
  parser.yyparse() ;
  in.close() ;
}

//----------------------------------------------------------------------//
// oRis link
//----------------------------------------------------------------------//

void FuzzyController::staticDestroy(Context* context)
{
  FuzzyController* controller = (FuzzyController *) context->getOwner() ;
  controller->destroy() ;
}

//----------------------------------------------------------------------//

void FuzzyController::staticCompile(Context* context)
{
  FuzzyController* controller = (FuzzyController *) context->getOwner() ;
  
  size_t i ;
  Strings srules ;
  const SubValueVect* rulesVect = controller->_rules->getValueVector() ;
  if (rulesVect->getSize()==0)
    // no rules declared
    return ;
  
  for (i=0; i<rulesVect->getSize(); i++)
    {
     Value* val = (*rulesVect)[i] ;
     srules += val->getStrings() + Strings("\n") ;
    }

  char * str=(char *)BlockAllocator::allocate(srules.len()+1);
  ::strcpy(str,srules);

  controller->compile(str, context->getInteger(1)) ;
  BlockAllocator::free(str);
}


//----------------------------------------------------------------------//

void FuzzyController::staticAddRulesFromFile(Context* context)
{
  FuzzyController* controller = (FuzzyController *) context->accessOwner() ;
  Strings fileName(getSimul()->locateFile(context->getStrings(0)));
  if(fileName.len())
    controller->addRulesFromFile(fileName);
  else
    {
    Strings message="Can't open file '"+
                    context->getStrings(0)+
                    "'";
    getSimul()->manageError(new Error(Error::RUNTIME,message));
    }
}




