%{
//-------------------------------------------------------------------------
//
// File    	:  fuzzyParse.y
// Description  :  Syntaxic analysis of fuzzy rules
//
// Author(s)   	: P. Reignier
// Date         : 01/04/98
// Copyright	: Laboratoire d'Informatique Industrielle (LI2)
//		  Ecole Nationale d'Ingenieurs de Brest (ENIB)
//		  CP 15 - 29608 Brest Cedex - FRANCE
//		  Tel: (33) 98.05.66.31, Fax: (33) 98.05.66.29
//		  e-mail: li2@li2.enib.fr
//-------------------------------------------------------------------------
// Copyright (c) 1996
// Laboratoire d'Informatique Industrielle (LI2)
// Ecole Nationale d'Ingenieurs de Brest (ENIB)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.
// ENIB/LI2 makes no representations about the suitability of this
// software for any purpose.
// It is provided "as is" without express or implied warranty.
//-------------------------------------------------------------------------

#include <iostream> 
#include <fstream>
#include <stdio.h>

#include "oRis/Main/errors.h"    
#include "oRis/Main/simul.h"
#include "fuzzyFNod.h"
#include "andConclFNod.h"   
#include "implyFNod.h"
#include "modFNod.h"
#include "varCondFNod.h"
#include "varConclFNod.h"
#include "andCondFNod.h"
#include "isAlsoFNod.h"
#include "orFNod.h"
#include "deltaFNod.h"
#include "isFNod.h"
#include "rulesBaseFNod.h"
#include "fuzzyFNod.h"
#include "linguisFNod.h"
#include "varConclFNod.h"

#include "Utility/strings.h"
#include "fuzzyController.h"

class FuzzyParser ;
#include "fuzzyScan.h"

#ifndef __MSDOS__
#define malloc(size) BlockAllocator::allocate(size);
#define free(ptr)    BlockAllocator::free(ptr);
#endif
%}

%name FuzzyParser
%define INHERIT : public BlockAllocatedObject

%define MEMBERS FuzzyLexer* _scanner ;\
	int _nbLine ;\
        Strings _line ; \
	FuzzyFNod** _ptrRulesTree ;\
        int _ruleNumber ; \
	void initParser(istream*, \
			FuzzyFNod**) ;\
        void incNbLine() ;\
        int getRulesNumber(void) const ; \
        virtual ~FuzzyParser(void) ;

%define CONSTRUCTOR_PARAM istream* stream, \
	  FuzzyFNod** ptrRulesTree
	 
%define CONSTRUCTOR_CODE initParser(stream, \
	ptrRulesTree)

%define LEX_BODY   {return _scanner->yylex(this) ;}

%start Program

%union {
  Strings* string ;   
  FuzzyFNod* node ;
}

%token IF IS THEN TERM AND OR 
%token PV V PO PF
%token DELTA

%type <string> TERM
%type <node> ListRules 
%type <node> ListActions
%type <node> VariableCond, VariableConcl
%type <node> ListModifiers
%type <node> TermComplex
%type <node> Rules
%type <node> Prop
%type <node> PropElem
%type <node> Action

%left OR
%left AND
%right MOD
%%

Program : DefRules
;

/*----------------------------------------------------------------------*/
/*                                                                      */
/* RULES DEFINITIONS                                                    */
/*                                                                      */
/*----------------------------------------------------------------------*/


DefRules :	ListRules 
                {
		  *_ptrRulesTree = $1 ;
		}
;


ListRules :	ListRules Rules 
                {
                  $$ = new RulesBaseFNod () ;
		  if ($1 != 0)
		    $$->insert($1) ;
		  $$->insert($2) ;
		}
                | /* empty */ {
		  $$ = 0 ;
		}
;


Rules : 	IF Prop THEN ListActions PV 
                {
		  $$ = new ImplyFNod(_ruleNumber) ;
		  $$->insert($2) ;
		  $$->insert($4) ;
		  _ruleNumber++ ;
		}
;


Prop :		PropElem 
                {
		  $$ = $1 ;
		}
                | PO Prop PF
                {
		  $$ = $2 ;
		}
		| Prop AND Prop 
                {
		  $$ = new AndCondFNod() ;
		  $$->insert($1) ;
		  $$->insert($3) ;
		}
		| Prop OR Prop 
                {
		  $$ = new OrFNod() ;
		  $$->insert($1) ;
		  $$->insert($3) ;
		}
;

PropElem :	VariableCond IS TermComplex 
                {
		  $$ = new IsFNod () ;
		  $$->insert($1) ;
		  $$->insert($3) ;
                }
;


TermComplex	: TERM
                {
		  // there is no modificator : it is just a linguistic term
		  $$ = new LinguisFNod(*$1) ;
		  delete $1 ;
		}
                | ListModifiers TERM
                {
		  LinguisFNod* lingFNod = new LinguisFNod(*$2) ;
		  delete $2 ;

		  // we have to reach the last modifier of $1 to 
		  // plug in the linguistic term
		  
		  FuzzyFNod* tmpFNod = $1 ;
		  while(tmpFNod->getNbFils() != 0)
		    tmpFNod = tmpFNod->getFils(0) ;
		  
		  // we now plugs-in the linguistic term
		  tmpFNod->insert(lingFNod) ;
		  $$ = $1 ;
		}
;

ListModifiers : ListModifiers TERM
              {
		// we have to go at the end of $1
		FuzzyFNod* tmpFNod = $1 ;
		while(tmpFNod->getNbFils() != 0)
		  tmpFNod = tmpFNod->getFils(0) ;

		// we now plugs-in the last modifier
		tmpFNod->insert(new ModFNod(*$2)) ;
		$$ = $1 ;
		delete $2 ;
	      }
              | TERM
              {
		$$ = new ModFNod(*$1) ;
		delete $1 ;
	      }

VariableCond : TERM 
             { 
	       $$ = new VarCondFNod(*$1) ;
	       delete $1 ;
	     }
             | DELTA TERM
             {
	       $$ = new DeltaFNod() ;
	       $$->insert(new VarCondFNod(*$2)) ;
	       delete $2 ;
	     }
; 

VariableConcl : TERM 
              { 
		$$ = new VarConclFNod(*$1) ;
		delete $1 ;
	      }
              | DELTA TERM
              {
		$$ = new DeltaFNod() ;
		$$->insert(new VarConclFNod(*$2)) ;
		delete $2 ;
	      }

ListActions :	Action {$$ = $1 ;}
                | ListActions AND ListActions 
                {
		  $$ = new AndConclFNod () ;
		  $$->insert($1) ;
		  $$->insert($3) ;
		}		  
;

Action : 	VariableConcl IS TermComplex 
                {
		  $$ = new IsAlsoFNod() ;
		  $$->insert($1) ;
		  $$->insert($3) ;
		}
;


%%

//----------------------------------------------------------------------//
//
// C++ CODE
//
//----------------------------------------------------------------------//

//------FuzzyParser::yyerror()------------------------------------------//

void FuzzyParser::yyerror (char *)
{
  char nb[100] ;
  sprintf(nb, "%d", _ruleNumber) ;
  Strings message = _line + "<--- : Syntax error rule nb " + nb ;
  getSimul()->manageError(new Error(Error::SYNTAX, message)); 
  return ;
}
		

//------FuzzyParser::incNbLine()---------------------------------------//

void FuzzyParser::incNbLine()
{
  _nbLine++ ;
}


//------FuzzyParser::initParser()---------------------------------------//

void FuzzyParser::initParser(istream *stream, 
				   FuzzyFNod** ptrRulesTree)
{
  _scanner=new FuzzyLexer(stream) ; 

  _line = Strings("") ;
  _nbLine = 1;
  _ruleNumber = 0;
  _ptrRulesTree = ptrRulesTree ;
}


//----------------------------------------------------------------------//

FuzzyParser::~FuzzyParser(void)
{
  delete _scanner ;
}

//----------------------------------------------------------------------//

int FuzzyParser::getRulesNumber(void) const 
{
  return _ruleNumber ;
}

