//-------------------------------------------------------------------------
//
// File         :  fuzzySet.cpp
// Description  :  fuzzy Set
//
// Author(s)    : P. Reignier
// Date         : 09/03/97
// Copyright    : Laboratoire d'Informatique Industrielle (LI2)
//                Ecole Nationale d'Ingenieurs de Brest (ENIB)
//                CP 15 - 29608 Brest Cedex - FRANCE
//                Tel: (33) 98.05.66.31, Fax: (33) 98.05.66.29
//                e-mail: li2@li2.enib.fr
//-------------------------------------------------------------------------
// Copyright (c) 1996
// Laboratoire d'Informatique Industrielle (LI2)
// Ecole Nationale d'Ingenieurs de Brest (ENIB)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.
// ENIB/LI2 makes no representations about the suitability of this
// software for any purpose.
// It is provided "as is" without express or implied warranty.
//-------------------------------------------------------------------------

#include <math.h>
#include <fstream>
#include <stdlib.h>
#include <assert.h>
#ifdef __MSDOS__
#include <windows.h>
#include <windef.h> // definition de min et max
#endif

#include "oRis/Main/agent.h"
#include "oRis/Main/model.h"
#include "Utility/stlVect.h"
#include "fuzzySet.h"

//----------------------------------------------------------------------//
//
//  Constructors
//
//----------------------------------------------------------------------//

//-----FuzzySet::initFuzzySet(double)-----------------------------------//

FuzzySet::FuzzySet (double value) 
  : BlockAllocatedObject()
{
  _value = value ;
  _type = CRISP ;
}


//-----FuzzySet::FuzzySet(StlVector<Point2D>)----------------------//

FuzzySet::FuzzySet (const StlVector<Point2D>& PointsInit) 
  : BlockAllocatedObject()
{
  _listePoints = PointsInit ;
  _type = FUZZY ;
}


//------FuzzySet::FuzzySet(FuzzySet&)------------------------------//

FuzzySet::FuzzySet(const FuzzySet& set) 
  : BlockAllocatedObject()
{
  _copy (set) ;
}


//----------------------------------------------------------------------//
//
// Destructors
//
//----------------------------------------------------------------------//


//-----FuzzySet::~FuzzySet() -------------------------------------------//

FuzzySet::~FuzzySet()
{
  _destroy () ;
}


//----------------------------------------------------------------------//
//
// Operators
//
//----------------------------------------------------------------------//


//------FuzzySet::operator=(FuzzySet&)----------------------------------//

FuzzySet& FuzzySet::operator= (const FuzzySet& set)
{
  if (this != &set)
  {
    _destroy () ;
    _copy (set) ;
  }

  return (*this) ;
}


//------operator<<()----------------------------------------------------//

ostream& operator<< (ostream& flot, const FuzzySet& set)
{
  if (set._type == FuzzySet::CRISP)
    flot << "crisp " << set._value ; 
  else 
    {
      flot << "[" ;
      for (unsigned int i=0; i != set._listePoints.size(); i++)
      {
        flot << "(" << set._listePoints[i] << ")" ;
        if (i!= set._listePoints.size()-1)
          flot << "," ;
      }
      flot << "]" ;
    }

  return (flot) ;
}


//----------------------------------------------------------------------
//
// Modifiers
//
//----------------------------------------------------------------------//


//------FuzzySet::setValue(f)-------------------------------------------//

void FuzzySet::setValue(double f) 
{
  _value = f ;
  _type = CRISP ;
  
  while (_listePoints.size() !=0)
    // on vide l'ensemble flou precedent
    _listePoints.pop_back() ;
}


//------FuzzySet::setValue(liste)---------------------------------------//

void FuzzySet::setValue(const StlVector<Point2D>& liste)
{
  _listePoints = liste ;
  _type = FuzzySet::FUZZY ;
}

//------FuzzySet::saturate(a)-------------------------------------------//

void FuzzySet::saturate(double a) 
{
  StlVector<Point2D> cstFunc ;
  cstFunc.push_back(Point2D(0,a)) ;
  FuzzySet cste(cstFunc) ;

  combineMin(cste) ;
}

//------FuzzySet::product(a)--------------------------------------------//

void FuzzySet::product(double a)
{
  unsigned int i ;
  double y ;

  if (_type == CRISP)
    // should never happen
    {
      _value *= a ;
      
      return ;
    }

  for (i=0; i<_listePoints.size(); i++)
    {
      y = _listePoints[i].getY() ;
      y *= a ;
      _listePoints[i].setY(y) ;
    }
}


//------FuzzySet::combineSum(FuzzySet&)---------------------------------//

void FuzzySet::combineSum(const FuzzySet& set) 
{
  Fusion tmp ;

#ifdef ASSERT
  assert((_type != CRISP) && (set.getType() != CRISP)) ;
#endif

  // we first add to both functions all the corresponding points
  _merge(_listePoints, set._listePoints, tmp) ;
  _intersect(tmp) ;

  _razListePoints() ;
  
  // we create the new function
  Fusion::const_iterator i ;
  for (i=tmp.begin(); i != tmp.end(); i++)
    {
      _listePoints.push_back(Point2D((*i).first, 
                                     (*i).second.first +
                                     (*i).second.second)) ;
    }

  simplify() ;

  // we do not normalize the set. Normalization must be done when all the
  // rules have been evaluated : else, the rule order is important.
}


//------FuzzySet::combineMax()------------------------------------------//

void FuzzySet::combineMax(const FuzzySet& set)
{
  Fusion tmp ;

#ifdef ASSERT
  assert((_type != CRISP) && (set.getType() != CRISP)) ;
#endif

  // we first add to both functions all the corresponding points
  _merge(_listePoints, set._listePoints, tmp) ;
  _intersect(tmp) ;

  _razListePoints() ;
  
  // we create the new function
  Fusion::const_iterator i ;
  for (i=tmp.begin(); i != tmp.end(); i++)
    {
      _listePoints.push_back(Point2D((*i).first, 
                                     max((*i).second.first,
                                         (*i).second.second))) ;
    }

  simplify() ;
}

//------FuzzySet::combineMin()------------------------------------------//

void FuzzySet::combineMin(const FuzzySet& set)
{
  Fusion tmp ;

#ifdef ASSERT
  assert((_type != CRISP) && (set.getType() != CRISP)) ;
#endif

  // we first add to both functions all the corresponding points
  _merge(_listePoints, set._listePoints, tmp) ;
  _intersect(tmp) ;

  _razListePoints() ;
  
  // we create the new function
  Fusion::const_iterator i ;
  for (i=tmp.begin(); i != tmp.end(); i++)
    {
      _listePoints.push_back(Point2D((*i).first, 
                                     min((*i).second.first,
                                         (*i).second.second))) ;
    }

  simplify() ;
}


//----------------------------------------------------------------------//
//
// Observers
// 
//----------------------------------------------------------------------//


//------FuzzySet::mu(double)--------------------------------------------//

double FuzzySet::mu (double x) const
{
  double y ;

  if (_type == CRISP)
    // we compare x to value
    {
      if (x == _value)
        y = 1.0 ;
      else
        y = 0.0 ;
    }
  else
    // we are in the fuzzy case
    y = _feval(_listePoints, x) ;
    
  return (y) ;
}


//------FuzzySet::integral()--------------------------------------------//

double FuzzySet::integral(void) const 
{
  double local_surface ;
  int i ;
  int capacity ;
  double surface ;
  double y_min ;
  double y_max ;
  double x_min ;
  double x_max ;

  if (_type == CRISP)
    return (0.0) ;

  capacity = _listePoints.size() ;

  if ((_listePoints[capacity-1].getY() >= Point2D::epsilon) ||
    (_listePoints[0].getY() >= Point2D::epsilon))
    // the surface is infinite
    return (HUGE_VAL) ;

  surface = 0.0 ;
  for (i=0; i<(capacity-1); i++)
    // compute the surface created by each segment
    {
      x_min = _listePoints[i].getX() ;
      x_max = _listePoints[i+1].getX() ;
      if ((x_max-x_min) >= Point2D::epsilon)
        // the surface is not 0
        {
          y_min = min(_listePoints[i].getY(), _listePoints[i+1].getY());
          y_max = max(_listePoints[i].getY(), _listePoints[i+1].getY());

          local_surface = (x_max - x_min) * y_min + 
            (x_max - x_min) * (y_max - y_min) / 2.0 ;
          surface += local_surface ;
        }
    }

  return (surface) ;
}



//------FuzzySet::xIntegral()-------------------------------------------//

double FuzzySet::xIntegral(void) const 
{
  // computes the integral of xf(x)

  double local_xsurface ;
  int i ;
  int capacity ;
  double xsurface ;
  double y0 ;
  double y1 ;
  double x0 ;
  double x1 ;

  if (_type == CRISP)
    return (0.0) ;

  capacity = _listePoints.size() ;

  if ((_listePoints[capacity-1].getY() >= Point2D::epsilon) ||
    (_listePoints[0].getY() >= Point2D::epsilon))
    // the surface is infinite
    return (HUGE_VAL) ;
  
  xsurface = 0.0 ;
  double a,b ;

  for (i=0; i<(capacity-1); i++)
    // compute the surface created by each segment
    {
      x0 = _listePoints[i].getX() ;
      x1 = _listePoints[i+1].getX() ;
      if ((x1-x0) >= Point2D::epsilon)
        // the surface is not 0
        {
          y0 = _listePoints[i].getY() ;
          y1 = _listePoints[i+1].getY() ;

          a = (y0-y1)/(x0-x1) ;
          b = (y1*x0 - x1*y0) / (x0-x1) ;

          local_xsurface = a/3.0*(pow(x1,3)-pow(x0,3)) + b/2.0*(x1*x1-x0*x0) ;
          xsurface += local_xsurface ;
        }
    }

  return (xsurface) ;
}


//------FuzzySet::maxMu()-----------------------------------------------//

bool FuzzySet::maxMu(StlVector<Point2D>& listeMax) const
{
  // returns the list of point where the function reaches its maximum value
  // returns false if the maximum is not reached at discrete points but on
  // a whole interval
  // maxMu always contains something valid at the end of the function

  double max ;
  unsigned int i ;

  if (_type == CRISP)
    // the maximum is reached on that point
    {
      listeMax.push_back(Point2D(_value,1)) ;
      return true ;
    }

  max = 0.0 ;

  for (i=0; i<_listePoints.size(); i++)
    if (_listePoints[i].getY() > max)
      max = _listePoints[i].getY() ;
  
  for (i=0; i<_listePoints.size(); i++)
    {
      if (_listePoints[i].getY() == max)
        {
          if ((i!=(_listePoints.size()-1)) && 
              (_listePoints[i+1].getY() == max))
            // we are on a flat part of the function => the maximum is not
            // reached on discrete points
            {
              listeMax.push_back(Point2D(_listePoints[i].getX(),max)) ;

              return (false) ;
            }
        }
    }

  for (i=0; i<_listePoints.size(); i++)
    if (_listePoints[i].getY() == max)
      listeMax.push_back(Point2D(_listePoints[i].getX(),max)) ;
  
  return (true) ;
}

//----------------------------------------------------------------------//

double FuzzySet::maxMin(const FuzzySet& set) const
{
  if ((_type != CRISP) && (set.getType() == CRISP))
    {
      double result ;
      result = set.getCrispValue() ;
      return mu(result) ;
    }
  
  if ((set.getType() != CRISP) && (_type == CRISP))
    {
      return set.mu(_value) ;
    }

  if ((set.getType() == CRISP) && (_type == CRISP))
    {
      double val2 ;
      val2 = set.getCrispValue() ;
      if (_value == val2)
        return 1.0 ;
      else
        return 0.0 ;
    }

  // both sets are fuzzy
  FuzzySet tmp(*this) ;
  tmp.combineMin(set) ;
  StlVector<Point2D> maxima ;
  tmp.maxMu(maxima) ;
  
  return maxima[0].getY();
}

//------FuzzySet::getCrispValue()---------------------------------------//

double FuzzySet::getCrispValue(void) const
{
  if (_type == CRISP)
      return _value ;
  else
    // we defuzzify
  {
    double integr = integral() ;
    double result ;
    if (integr != 0)
      result = xIntegral() / integr ;
    else
      // arbitrary value 
      result = 0.0 ;
    return result ;
  }
}


//------FuzzySet::getFuzzyValue()--------------------------------------//

const StlVector<Point2D>& FuzzySet::getFuzzyValue(void) const
{
#ifdef ASSERT
  assert(_type == FUZZY) ;
#endif

  return _listePoints ;
}


//-----FuzzySet::getValue(i,point)--------------------------------------//

const Point2D& FuzzySet::getValue(int i) const
{
#ifdef ASSERT
  assert(_type == FUZZY) ;
#endif

  return _listePoints[i] ;
}


//------FuzzySet::getSize()---------------------------------------------//

int FuzzySet::getSize() const
{
  if (_type == CRISP)
    return (1) ;
  else
    return (_listePoints.size()) ;
}

//----------------------------------------------------------------------//
//
// Utilities
//
//----------------------------------------------------------------------//


//------FuzzySet::_copy(const FuzzySet&)--------------------------------//

void FuzzySet::_copy(const FuzzySet& set)
{
  _type  = set._type ;

  if (_type == CRISP)
    _value = set._value ;
  else
    _listePoints = set._listePoints ;
}


//------FuzzySet::_destroy()--------------------------------------------//

void FuzzySet::_destroy()
{}


//----------------------------------------------------------------------

double FuzzySet::_feval (const StlVector<Point2D>& liste, double x) const
{
  double y ;
  int i ;
  int capacity ;
  bool found ;

  capacity = liste.size() ;

  if (x <= liste[0].getX())
    // if x is outside on the left of the list of points, 
    // the value of the function is given by the most left point.
    y = liste[0].getY() ;
  else if (x >= liste[capacity-1].getX())
    // if x is outside on the right of the list of points, 
    // the value of the function is given by the most right point.
    y = liste[capacity-1].getY() ;
  else
    // we find the segment corresponding to the x value
    {
      for (i=0, found=false; ((i <(capacity-1)) && !found); i++)
        found = ((x >= liste[i].getX()) &&
                 (x <= liste[i+1].getX())) ;
      i-- ;
      if (liste[i].getY() == liste[i+1].getY())
        // the function is constant on the segment
        y = liste[i].getY() ;
      else if (liste[i].getX() == liste[i+1].getX())
        // the segment is limited to a point
        y = liste[i].getY() ;
      else
        // this is a standart segment
        y = (x-liste[i].getX())/(liste[i+1].getX()-liste[i].getX()) *
          (liste[i+1].getY()-liste[i].getY()) + liste[i].getY() ;
    }

  return (y) ;
}

//----------------------------------------------------------------------//

void FuzzySet::_merge(const StlVector<Point2D>& vec1, 
                      const StlVector<Point2D>& vec2,
                      Fusion& res) const
{
  StlVector<Point2D>::const_iterator i ;

  // on cree une liste triee comprenant tous les points remarquables
  // du premier et du second vecteur. Les informations stockees dans 
  // la map sont une paire dont le premier element est la coordonnee
  // y de vec1 pour le point considere et le deuxieme element est l'ordonnee
  // de vec2

  for (i=vec1.begin(); i!=vec1.end(); i++)
    res.insert(Fusion::value_type((*i).getX(),
                                  StlPair<double,double>((*i).getY(),
                                                         _feval(vec2,(*i).getX())
                                                         )
                                  )
               );

  for (i=vec2.begin(); i!=vec2.end(); i++)
    res.insert(Fusion::value_type((*i).getX(),
                                  StlPair<double,double>(_feval(vec1,(*i).getX()),
                                                         (*i).getY()
                                                         )
                                  )
               );
}

//----------------------------------------------------------------------//

void FuzzySet::_intersect(Fusion& fonction) const
{
  // complete the given two functions with their respective
  // intersection points
  
  Fusion::const_iterator i ;

  if (fonction.size() == 1)
    // both functions are constant => there are either no or 
    // an infinity of intersection points
    return ;
  
  i = fonction.begin() ;
  bool finished = false ;

  while (!finished)
    {
      Fusion::const_iterator current, next ;
      current = i ;
      next = ++i ;
      if (next == fonction.end())
        finished = true ;
      else
        {
          double y1_current, y2_current ;
          double y1_next, y2_next ;
          
          y1_current = (*current).second.first ;
          y2_current = (*current).second.second ;

          y1_next = (*next).second.first ;
          y2_next = (*next).second.second ;

          double signe1 = y2_current-y1_current ;
          double signe2 = y2_next-y1_next ;
          if (signe1*signe2 < 0.0)
            // there is an intersection
            {
              double delta_x = (*next).first-(*current).first ;
              double delta_y = (*next).second.second-(*current).second.second ;
              double a0 = delta_y/delta_x ;
              double b0 = (*current).second.second - 
                delta_y/delta_x * (*current).first ;

              delta_y = (*next).second.first-(*current).second.first;
              double a1 = delta_y/delta_x ;
              double b1 = (*current).second.first - 
                delta_y/delta_x * (*current).first ;

              double x = (b1-b0)/(a0-a1) ;
              double y = a0*x+b0 ;
              
              // we add the intersection point
              fonction.insert(Fusion::value_type(x,
                                                 StlPair<double,double>(y,y)));
            }
        }
    }
}

//----------------------------------------------------------------------//

void FuzzySet::simplify(void) 
{
  StlVector<Point2D> tmp ;
  StlVector<Point2D>::iterator i ;
  Point2D currentPoint ;
  Point2D firstPoint ;

  i = _listePoints.begin() ;

  while(i != _listePoints.end())
    {
      currentPoint = firstPoint = (*i++) ;
      tmp.push_back(currentPoint) ;
      if (i != _listePoints.end())
        {
          if ((*i).getY() == currentPoint.getY())
            // there is some doublons
            {
              while((i != _listePoints.end()) &&
                    ((*i).getY() == currentPoint.getY())
                    )
                currentPoint = (*i++) ;
              if (firstPoint != currentPoint)
                tmp.push_back(currentPoint) ;
            }
        }
    }

  _listePoints = tmp ;
}

//----------------------------------------------------------------------//

void FuzzySet::normalize(void)
{
  StlVector<Point2D>::iterator i ;
  double max=0 ;
  double y ;

  for (i=_listePoints.begin(); i!=_listePoints.end(); i++)
    {
      if ((*i).getY() > max)
        max = (*i).getY() ;
    }

  if (max > 1.0)
    // we need to normalize to bring the set back to one
    {
      for (i=_listePoints.begin(); i!=_listePoints.end(); i++)
        {
          y = (*i).getY() ;
          (*i).setY(y/max) ;
        }
    }
}

//----------------------------------------------------------------------//

FuzzySet FuzzySet::fNot(void) const
{
  FuzzySet result ;

  if (getType() == FuzzySet::CRISP)
    {
      double value ;
      value=getCrispValue() ;
      result.setValue(1-value) ;
    }
  else
    {
      StlVector<Point2D> newSet ;
      newSet = getFuzzyValue() ;

      StlVector<Point2D>::iterator i ;
      for (i=newSet.begin(); i != newSet.end(); i++)
        (*i).setY(1-(*i).getY()) ;
      result.setValue(newSet) ;
    }

  return result ;
}

//----------------------------------------------------------------------//

FuzzySet FuzzySet::fMoreOrLess(void) const
{
  FuzzySet result ;

  if (getType() == FuzzySet::CRISP)
    {
      double value ;
      value = getCrispValue() ;
      result.setValue(pow(value,0.5)) ;
    }

  // we create intermediate points to be able to bend
  // more efficiently the function
  StlVector<Point2D> newSet ;
  const StlVector<Point2D>& oldSet = getFuzzyValue() ;

  Point2D point ;
  Point2D intermediate ;

  if (oldSet.size() == 1)
    // just one point on the set : we cannot create intermediate points
    {
      newSet.push_back(Point2D(oldSet[0].getX(), 
                               pow(oldSet[0].getY(),0.5))) ;
      result.setValue(newSet) ;
    }
  else
    {
      size_t i ;
      double deltaX ;
      double x,y ;
      for (i = 0; i < (oldSet.size()-1); i++) 
        {
          point.setX(oldSet[i].getX()) ;
          point.setY(pow(oldSet[i].getY(),0.5)) ;
          newSet.push_back(point) ;
          
          deltaX = oldSet[i+1].getX() - oldSet[i].getX() ;

          x = oldSet[i].getX() + deltaX / 3.0 ;
          y = mu(x) ;
          intermediate.setX(x) ;
          intermediate.setY(pow(y,0.5)) ;
          newSet.push_back(intermediate) ;

          x = oldSet[i].getX() + 2.0 * deltaX / 3.0 ;
          y = mu(x) ;
          intermediate.setX(x) ;
          intermediate.setY(pow(y,0.5)) ;
          newSet.push_back(intermediate) ;
        }

      // we add the last point of the original set
      point.setX(oldSet[i].getX()) ;
      point.setY(pow(oldSet[i].getY(),0.5)) ;
      newSet.push_back(point) ;

      result.setValue(newSet) ;
      result.simplify() ;
    }

  return result ;
}

//----------------------------------------------------------------------//

FuzzySet FuzzySet::fVery(void) const
{
  FuzzySet result ;

  if (getType() == FuzzySet::CRISP)
    {
      double value ;
      value = getCrispValue() ;
      result.setValue(pow(value,2)) ;
    }

  // we create intermediate points to be able to bend
  // more efficiently the function
  StlVector<Point2D> newSet ;
  const StlVector<Point2D>& oldSet = getFuzzyValue() ;

  Point2D point ;
  Point2D intermediate ;

  if (oldSet.size() == 1)
    // just one point on the set : we cannot create intermediate points
    {
      newSet.push_back(Point2D(oldSet[0].getX(), 
                               pow(oldSet[0].getY(),2))) ;
      result.setValue(newSet) ;
    }
  else
    {
      size_t i ;
      double deltaX ;
      double x,y ;
      for (i = 0; i < (oldSet.size()-1); i++) 
        {
          point.setX(oldSet[i].getX()) ;
          point.setY(pow(oldSet[i].getY(),2)) ;
          newSet.push_back(point) ;
          
          deltaX = oldSet[i+1].getX() - oldSet[i].getX() ;

          x = oldSet[i].getX() + deltaX / 3.0 ;
          y = mu(x) ;
          intermediate.setX(x) ;
          intermediate.setY(pow(y,2)) ;
          newSet.push_back(intermediate) ;

          x = oldSet[i].getX() + 2.0 * deltaX / 3.0 ;
          y = mu(x) ;
          intermediate.setX(x) ;
          intermediate.setY(pow(y,2)) ;
          newSet.push_back(intermediate) ;
        }

      // we add the last point of the original set
      point.setX(oldSet[i].getX()) ;
      point.setY(pow(oldSet[i].getY(),2)) ;
      newSet.push_back(point) ;

      result.setValue(newSet) ;
      result.simplify() ;
    }

  return result ;
}
